import os
import unittest
from copy import deepcopy
from tempfile import TemporaryDirectory

from dhpython.depends import Dependencies
from dhpython.fs import ScanResult
from dhpython.options import DHPythonOptions
from dhpython.pydist import Standard

from .test_depends import prime_pydist, pydist


class TestDistutilsExtra(unittest.TestCase):
    options = DHPythonOptions(guess_deps=True)
    pydist = {
        "bar": [pydist(name="bar", dependency="python3-bar")],
        "baz": [pydist(name="baz", dependency="python3-baz", standard=Standard.PEP386)],
        "quux": [
            pydist(name="quux", dependency="python3-quux", standard=Standard.PEP386)
        ],
    }
    pkg = "foo"
    impl = "cpython3"
    stats = ScanResult(
        {
            "compile": False,
            "dist-info": set(),
            "egg-info": {"PKG-INFO"},
            "ext_no_version": set(),
            "ext_stableabi": set(),
            "ext_vers": set(),
            "nsp.txt": set(),
            "private_dirs": {},
            "public_vers": set(),
            "requires.txt": set(),
            "shebangs": set(),
        }
    )

    def test_depends_on_bar(self) -> None:
        self.d = Dependencies(self.pkg, self.impl)
        stats = deepcopy(self.stats)
        self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tempdir.cleanup)
        old_wd = os.getcwd()
        os.chdir(self.tempdir.name)
        self.addCleanup(os.chdir, old_wd)
        with open(self.tempdir.name + "/PKG-INFO", "w", encoding="UTF-8") as f:
            f.write("""\
Metadata-Version: 2.1
Name: gTranscribe
Version: 0.11
Summary: gTranscribe
Home-page: https://github.com/innir/gtranscribe
Author: Philip Rinn
Author-email: rinni@inventati.org
License: GPL-3
Requires: bar

gTranscribe is a software focused on easy transcription of spoken words.
""")
        cleanup = prime_pydist(self.impl, self.pydist)
        self.addCleanup(cleanup)
        self.d.parse(stats, self.options)
        self.assertIn("python3-bar", self.d.depends)



# Copyright © 2022 Stefano Rivera <stefanor@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Handle Environment Markers
https://www.python.org/dev/peps/pep-0508/#environment-markers

TODO: Ideally replace with the packaging library, but the API is currently
private: https://github.com/pypa/packaging/issues/496
"""

import re
from typing import Literal, NamedTuple, cast

SIMPLE_ENV_MARKER_RE = re.compile(
    r"""
    (?P<marker>[a-z_]+)
    \s*
    (?P<op><=?|>=?|[=!~]=|===)
    \s*
    (?P<quote>['"])
    (?P<value>.*)  # Could contain additional markers
    (?P=quote)
    """,
    re.VERBOSE,
)
COMPLEX_ENV_MARKER_RE = re.compile(
    r"""
    (?:\s|\))
    (?:and|or)
    (?:\s|\()
    """,
    re.VERBOSE,
)


class ComplexEnvironmentMarker(Exception):
    pass


OP = Literal["<", ">", "<=", ">=", "==", "!=", "~=", "==="]


class ParsedMarker(NamedTuple):
    marker: str
    op: OP
    value: str


def parse_environment_marker(marker: str) -> ParsedMarker:
    """Parse a simple marker of <= 1 environment restriction"""
    marker = marker.strip()
    if marker.startswith("(") and marker.endswith(")"):
        marker = marker[1:-1].strip()

    m = COMPLEX_ENV_MARKER_RE.search(marker)
    if m:
        raise ComplexEnvironmentMarker()

    m = SIMPLE_ENV_MARKER_RE.match(marker)
    if not m:
        raise ComplexEnvironmentMarker()

    return ParsedMarker(
        marker=m.group("marker"),
        op=cast(OP, m.group("op")),
        value=m.group("value"),
    )



# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from functools import partial
from os.path import exists, join

from dhpython import PKG_PREFIX_MAP, MINPYCDEP
from dhpython.fs import ScanResult
from dhpython.debhelper import BD, DebHelper
from dhpython.options import DHPythonOptions
from dhpython.pydist import (
    NewDependencies,
    guess_dependency,
    parse_pydep,
    parse_requires_dist,
)
from dhpython.version import default, supported, Version, VersionRange

log = logging.getLogger("dhpython")


class Dependencies:
    """Store relations (dependencies, etc.) between packages."""

    impl: str
    package: str
    bdep: BD | None
    is_debug_package: bool
    ipkg_vtpl: str
    ipkg_tpl: str
    ipkg_tpl_ma: str
    ipkg_vtpl_ma: str
    python_dv_in_bd: bool
    depends: set[str]
    recommends: list[str]
    suggests: list[str]
    enhances: list[str]
    breaks: list[str]
    rtscripts: list[tuple[str, str]]

    def __init__(
        self,
        package: str,
        impl: str = "cpython3",
        bdep: BD | None = None,
    ) -> None:
        self.impl = impl
        self.package = package
        bdep = self.bdep = bdep or {}
        self.is_debug_package = dbgpkg = package.endswith("-dbg")

        self.ipkg_vtpl = "python%s-dbg" if dbgpkg else "python%s"
        self.ipkg_tpl = "python3-dbg" if dbgpkg else "python3"
        self.ipkg_tpl_ma = self.ipkg_tpl + ":any"
        self.ipkg_vtpl_ma = self.ipkg_vtpl + ":any"

        self.python_dev_in_bd = (
            "python-dev" in bdep
            or "python-all-dev" in bdep
            or "python3-dev" in bdep
            or "python3-all-dev" in bdep
        )

        self.depends = set()
        self.recommends = []
        self.suggests = []
        self.enhances = []
        self.breaks = []
        self.rtscripts = []

    def export_to(self, dh: DebHelper) -> None:
        """Fill in debhelper's substvars."""
        prefix = PKG_PREFIX_MAP.get(self.impl, "misc")
        for i in sorted(self.depends):
            dh.addsubstvar(self.package, f"{prefix}:Depends", i)
        for i in sorted(self.recommends):
            dh.addsubstvar(self.package, f"{prefix}:Recommends", i)
        for i in sorted(self.suggests):
            dh.addsubstvar(self.package, f"{prefix}:Suggests", i)
        for i in sorted(self.enhances):
            dh.addsubstvar(self.package, f"{prefix}:Enhances", i)
        for i in sorted(self.breaks):
            dh.addsubstvar(self.package, f"{prefix}:Breaks", i)
        for script in sorted(self.rtscripts):
            dh.add_rtupdate(self.package, script)

    def __str__(self) -> str:
        return "D={}; R={}; S={}; E={}, B={}; RT={}".format(
            self.depends,
            self.recommends,
            self.suggests,
            self.enhances,
            self.breaks,
            self.rtscripts,
        )

    def depend(self, value: str | None) -> None:
        if value and value not in self.depends:
            self.depends.add(value)

    def recommend(self, value: str | None) -> None:
        if value and value not in self.recommends:
            self.recommends.append(value)

    def suggest(self, value: str | None) -> None:
        if value and value not in self.suggests:
            self.suggests.append(value)

    def enhance(self, value: str | None) -> None:
        if value and value not in self.enhances:
            self.enhances.append(value)

    def break_(self, value: str | None) -> None:
        if value and value not in self.breaks:
            self.breaks.append(value)

    def rtscript(self, value: tuple[str, str]) -> None:
        if value not in self.rtscripts:
            self.rtscripts.append(value)

    def add_new_dependencies(self, new_dependencies: NewDependencies) -> None:
        for dep in new_dependencies["depends"]:
            self.depend(dep)
        for dep in new_dependencies["recommends"]:
            self.recommend(dep)
        for dep in new_dependencies["suggests"]:
            self.suggest(dep)

    def parse(self, stats: ScanResult, options: DHPythonOptions) -> None:
        log.debug("generating dependencies for package %s", self.package)
        tpl = self.ipkg_tpl
        vtpl = self.ipkg_vtpl
        tpl_ma = self.ipkg_tpl_ma
        vtpl_ma = self.ipkg_vtpl_ma
        vrange = options.vrange

        if vrange and any(
            (
                stats["compile"],
                stats["public_vers"],
                stats["ext_vers"],
                stats["ext_no_version"],
                stats["ext_stableabi"],
                stats["shebangs"],
            )
        ):
            if any((stats["compile"], stats["public_vers"], stats["shebangs"])):
                tpl_tmp = tpl_ma
            else:
                tpl_tmp = tpl
            minv = vrange.minver
            # note it's an open interval (i.e. do not add 1 here!):
            maxv = vrange.maxver
            if minv == maxv:
                self.depend(vtpl % minv)
                minv = maxv = None
            if minv:
                self.depend(f"{tpl_tmp} (>= {minv}~)")
            if maxv:
                self.depend(f"{tpl_tmp} (<< {maxv})")

        if stats["ext_vers"]:
            sorted_vers = sorted(stats["ext_vers"])
            minv = sorted_vers[0]
            maxv = sorted_vers[-1]
            # self.depend('|'.join(vtpl % i for i in stats['ext_vers']))
            self.depend(f"{tpl} (>= {minv}~)")
            self.depend(f"{tpl} (<< {maxv + 1})")
        elif stats["ext_stableabi"]:
            self.depend("%s" % tpl)

        # make sure py{,3}compile binary is available
        if stats["compile"] and self.impl in MINPYCDEP:
            self.depend(MINPYCDEP[self.impl])

        for ipreter in stats["shebangs"]:
            self.depend("%s:any" % ipreter)

        supported_versions = supported(self.impl)
        default_version = default(self.impl)
        for private_dir, details in stats["private_dirs"].items():
            versions = list(
                i.version
                for i in details.get("shebangs", [])
                if i.version and i.version.minor
            )

            for v in versions:
                if v in supported_versions:
                    self.depend(vtpl_ma % v)
                else:
                    log.info(
                        "dependency on %s (from shebang) ignored"
                        " - it's not supported anymore",
                        vtpl % v,
                    )
            # /usr/bin/python{,3} shebang → add python{,3} to Depends
            if any(
                True
                for i in details.get("shebangs", [])
                if i.version is None or i.version.minor is None
            ):
                self.depend(tpl_ma)

            extensions: list[Version] = []
            if self.python_dev_in_bd:
                extensions = sorted(details.get("ext_vers", set()))
                # self.depend('|'.join(vtpl % i for i in extensions))
                if extensions:
                    self.depend(f"{tpl} (>= {extensions[0]}~)")
                    self.depend(f"{tpl} (<< {extensions[-1] + 1})")
                elif details.get("ext_no_version"):
                    # assume unrecognized extension was built for default interpreter version
                    self.depend(f"{tpl} (>= {default_version}~)")
                    self.depend(f"{tpl} (<< {default_version + 1})")

            if details.get("compile"):
                if self.impl in MINPYCDEP:
                    self.depend(MINPYCDEP[self.impl])
                args = ""
                if extensions:
                    args += "-V %s" % VersionRange(
                        minver=extensions[0], maxver=extensions[-1]
                    )
                elif len(versions) == 1:  # only one version from shebang
                    # if versions[0] in supported_versions:
                    args += "-V %s" % versions[0]
                    # ... otherwise compile with default version
                elif details.get("ext_no_version"):
                    # assume unrecognized extension was built for default interpreter version
                    args += "-V %s" % default_version
                elif vrange:
                    args += "-V %s" % vrange
                    if vrange.minver == vrange.maxver:
                        self.depend(vtpl % vrange.minver)
                    else:
                        if vrange.minver:  # minimum version specified
                            self.depend(f"{tpl_ma} (>= {vrange.minver}~)")
                        if vrange.maxver:  # maximum version specified
                            self.depend(f"{tpl_ma} (<< {vrange.maxver + 1})")

                for regex in options.regexpr or []:
                    args += " -X '%s'" % regex.pattern.replace("'", r"'\''")
                self.rtscript((private_dir, args))

        guess_deps = partial(
            guess_dependency,
            impl=self.impl,
            bdep=self.bdep,
            accept_upstream_versions=options.accept_upstream_versions,
        )
        if options.guess_deps:
            for fn in stats["requires.txt"]:
                # TODO: should options.recommends and options.suggests be
                # removed from requires.txt?
                self.add_new_dependencies(
                    parse_pydep(
                        self.impl,
                        fn,
                        bdep=self.bdep,
                        depends_sec=options.depends_section,
                        recommends_sec=options.recommends_section,
                        suggests_sec=options.suggests_section,
                    )
                )
            for fpath in stats["egg-info"]:
                with open(fpath, encoding="utf-8") as fp:
                    for line in fp:
                        if line.startswith("Requires: "):
                            req = line[10:].strip()
                            self.depend(guess_deps(req=req))
            for fpath in stats["dist-info"]:
                self.add_new_dependencies(
                    parse_requires_dist(
                        self.impl,
                        fpath,
                        bdep=self.bdep,
                        depends_sec=options.depends_section,
                        recommends_sec=options.recommends_section,
                        suggests_sec=options.suggests_section,
                    )
                )

        # add dependencies from --depends
        for item in options.depends or []:
            self.depend(guess_deps(req=item))
        # add dependencies from --recommends
        for item in options.recommends or []:
            self.recommend(guess_deps(req=item))
        # add dependencies from --suggests
        for item in options.suggests or []:
            self.suggest(guess_deps(req=item))
        # add dependencies from --requires
        for fn in options.requires or []:
            fpath = join("debian", self.package, fn)
            if not exists(fpath):
                fpath = fn
                if not exists(fpath):
                    log.warning("cannot find requirements file: %s", fn)
                    continue
            self.add_new_dependencies(
                parse_pydep(
                    self.impl,
                    fpath,
                    bdep=self.bdep,
                    depends_sec=options.depends_section,
                    recommends_sec=options.recommends_section,
                    suggests_sec=options.suggests_section,
                )
            )

        log.debug(self)



from setuptools import setup, Extension

setup_args = dict(
    ext_modules=[
        Extension(
            'foo.ext',
            ['foo/ext.c'],
            depends=['foo/ext.h'],
            py_limited_api=True,
        )
    ],
    headers=['foo/ext.h'],
)
setup(**setup_args)



# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
from argparse import ArgumentParser, SUPPRESS
from dataclasses import dataclass, field
from os import environ

from dhpython.version import VersionRange


@dataclass
class DHPythonOptions:
    O: list[str] = field(default_factory=list)
    accept_upstream_versions: bool = False
    arch: bool | None = None
    clean_dbg_pkg: bool = True
    compile_all: bool = False
    depends: list[str] = field(default_factory=list)
    depends_section: list[str] = field(default_factory=list)
    guess_deps: bool = True
    ignore_shebangs: bool = False
    no_ext_rename: bool = False
    no_package: list[str] = field(default_factory=list)
    no_shebang_rewrite: bool = False
    package: list[str] = field(default_factory=list)
    private_dir: str | None = None
    recommends: list[str] = field(default_factory=list)
    recommends_section: list[str] = field(default_factory=list)
    regexpr: list[re.Pattern[str]] = field(default_factory=list)
    remaining_packages: bool = False
    requires: list[str] = field(default_factory=list)
    shebang: str | None = None
    skip_private: bool = False
    suggests: list[str] = field(default_factory=list)
    suggests_section: list[str] = field(default_factory=list)
    verbose: bool = environ.get("DH_VERBOSE") == "1"
    vrange: VersionRange | None = None
    write_log: bool = False


def compiled_regex(string: str) -> re.Pattern[str]:
    """argparse regex type"""
    try:
        return re.compile(string)
    except re.error:
        raise ValueError("regular expression is not valid")


def build_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument("--version", action="version", version="%(prog)s DEVELV")
    parser.add_argument(
        "--no-guessing-deps",
        action="store_false",
        dest="guess_deps",
        help="disable guessing dependencies",
    )
    parser.add_argument(
        "--skip-private", action="store_true", help="don't check private directories"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="turn verbose mode on",
    )
    # arch=False->arch:all only, arch=True->arch:any only, None->all of them
    parser.add_argument(
        "-i",
        "--indep",
        action="store_false",
        dest="arch",
        default=None,
        help="act on architecture independent packages",
    )
    parser.add_argument(
        "-a",
        "-s",
        "--arch",
        action="store_true",
        dest="arch",
        help="act on architecture dependent packages",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_false", dest="verbose", help="be quiet"
    )
    parser.add_argument(
        "-p",
        "--package",
        action="append",
        metavar="PACKAGE",
        help="act on the package named PACKAGE",
    )
    parser.add_argument(
        "-N",
        "--no-package",
        action="append",
        metavar="PACKAGE",
        help="do not act on the specified package",
    )
    parser.add_argument(
        "--remaining-packages",
        action="store_true",
        help="Do not act on the packages which have already been acted on by "
        "this debhelper command earlier",
    )
    parser.add_argument(
        "--compile-all",
        action="store_true",
        help="compile all files from given private directory in postinst, not "
        "just the ones provided by the package",
    )
    parser.add_argument(
        "-V",
        type=VersionRange,
        dest="vrange",
        metavar="[X.Y][-][A.B]",
        help="specify list of supported Python versions. See py3compile(1) for "
        "examples",
    )
    parser.add_argument(
        "-X",
        "--exclude",
        action="append",
        dest="regexpr",
        type=compiled_regex,
        metavar="REGEXPR",
        help="exclude .py files that match given REGEXPR from "
        "byte-compilation, in a private dir. You may use this option "
        "multiple times to build up a list of things to exclude.",
    )
    parser.add_argument(
        "--accept-upstream-versions",
        action="store_true",
        help="accept upstream versions while translating Python dependencies "
        "into Debian ones",
    )
    parser.add_argument(
        "--depends",
        action="append",
        metavar="REQ",
        help="translate given requirements into Debian dependencies and add "
        "them to ${python3:Depends}. Use it for missing items in "
        "requires.txt.",
    )
    parser.add_argument(
        "--depends-section",
        action="append",
        metavar="SECTION",
        help="translate requirements from given section into Debian "
        "dependencies and add them to ${python3:Depends}",
    )
    parser.add_argument(
        "--recommends",
        action="append",
        metavar="REQ",
        help="translate given requirements into Debian dependencies and add "
        "them to ${python3:Recommends}",
    )
    parser.add_argument(
        "--recommends-section",
        action="append",
        metavar="SECTION",
        help="translate requirements from given section into Debian "
        "dependencies and add them to ${python3:Recommends}",
    )
    parser.add_argument(
        "--suggests",
        action="append",
        metavar="REQ",
        help="translate given requirements into Debian dependencies and add "
        "them to ${python3:Suggests}",
    )
    parser.add_argument(
        "--suggests-section",
        action="append",
        metavar="SECTION",
        help="translate requirements from given section into Debian "
        "dependencies and add them to ${python3:Suggests}",
    )
    parser.add_argument(
        "--requires",
        action="append",
        metavar="FILE",
        help="translate requirements from given file into Debian dependencies "
        "and add them to ${python3:Depends}",
    )
    parser.add_argument(
        "--shebang", metavar="COMMAND", help="use given command as shebang in scripts"
    )
    parser.add_argument(
        "--ignore-shebangs",
        action="store_true",
        help="do not translate shebangs into Debian dependencies",
    )
    parser.add_argument(
        "--no-dbg-cleaning",
        action="store_false",
        dest="clean_dbg_pkg",
        help="do not remove files from debug packages",
    )
    parser.add_argument(
        "--no-ext-rename",
        action="store_true",
        help="do not add magic tags nor multiarch tuples to extension file " "names)",
    )
    parser.add_argument(
        "--no-shebang-rewrite", action="store_true", help="do not rewrite shebangs"
    )
    parser.add_argument(
        "private_dir",
        nargs="?",
        help="Private directory containing Python modules (optional)",
    )
    # debhelper options:
    parser.add_argument("-O", action="append", help=SUPPRESS)
    return parser



import os
import logging
import platform
import unittest
from collections.abc import Callable, Sequence
from tempfile import TemporaryDirectory
from typing import Any
from unittest.mock import patch

from dhpython.fs import ScanResult
from dhpython.pydist import PyDist, Standard
from dhpython.depends import Dependencies
from dhpython.options import DHPythonOptions
from dhpython.version import Version


def pep386(d: dict[str, str]) -> dict[str, list[PyDist]]:
    """Mark all pydist entries as being PEP386"""
    result: dict[str, list[PyDist]] = {}
    for name, dependency in d.items():
        result[name] = [
            pydist(name=name, dependency=dependency, standard=Standard.PEP386)
        ]
    return result


def pydist(
    name: str,
    dependency: str,
    versions: set[Version] = set(),
    standard: Standard | None = None,
    rules: list[str] | None = None,
) -> PyDist:
    return PyDist(
        name=name,
        versions=versions,
        dependency=dependency,
        standard=standard,
        rules=rules or [],
    )


def prime_pydist(impl: str, contents: dict[str, list[PyDist]]) -> Callable[[], None]:
    """Fake the pydist data for impl. Returns a cleanup function"""
    from dhpython.pydist import normalize_name

    normalized_pydist = {}
    for name, entries in contents.items():
        normalized_pydist[normalize_name(name)] = entries

    def fake(req_impl: str) -> dict[str, list[PyDist]]:
        if req_impl == impl:
            return normalized_pydist
        return {}

    patcher = patch("dhpython.pydist.load", side_effect=fake)
    patcher.start()
    return patcher.stop


class DependenciesTestCase(unittest.TestCase):
    pkg = "foo"
    impl = "cpython3"
    pydist: dict[str, list[PyDist]] = {}
    stats: dict[str, Any] = {}
    prepared_stats: ScanResult
    requires: dict[str, Sequence[str]] = {}
    dist_info_metadata: dict[str, Sequence[str]] = {}
    options = DHPythonOptions()
    parse = True

    def setUp(self) -> None:
        self.d = Dependencies(self.pkg, self.impl)

        stats = ScanResult(
            {
                "compile": False,
                "dist-info": set(),
                "egg-info": set(),
                "ext_no_version": set(),
                "ext_stableabi": set(),
                "ext_vers": set(),
                "nsp.txt": set(),
                "private_dirs": {},
                "public_vers": set(),
                "requires.txt": set(),
                "shebangs": set(),
            }
        )
        for k, v in self.stats.items():
            assert k in stats
            stats[k] = v  # type: ignore

        write_files = {}
        if self.requires:
            for fn, lines in self.requires.items():
                write_files[fn] = lines
                stats["requires.txt"].add(fn)

        if self.dist_info_metadata:
            for fn, lines in self.dist_info_metadata.items():
                write_files[fn] = lines
                stats["dist-info"].add(fn)

        if write_files:
            self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
            self.addCleanup(self.tempdir.cleanup)
            old_wd = os.getcwd()
            os.chdir(self.tempdir.name)
            self.addCleanup(os.chdir, old_wd)

        for fn, lines in write_files.items():
            os.makedirs(os.path.dirname(fn))
            with open(fn, "w", encoding="UTF-8") as f:
                f.write("\n".join(lines))

        cleanup = prime_pydist(self.impl, self.pydist)
        self.addCleanup(cleanup)

        if self.parse:
            self.d.parse(stats, self.options)
        else:
            self.prepared_stats = stats

    def assertNotInDepends(self, pkg: str) -> None:
        """Assert that pkg doesn't appear *anywhere* in self.d.depends"""
        for dep in self.d.depends:
            for alt in dep.split("|"):
                alt = alt.strip().split("(", 1)[0].strip()
                if pkg == alt:
                    raise AssertionError(f"{pkg} appears in {alt}")


class TestRequiresCPython3(DependenciesTestCase):
    options = DHPythonOptions(guess_deps=True)
    pydist = {
        "bar": [pydist(name="bar", dependency="python3-bar")],
        "baz": [pydist(name="baz", dependency="python3-baz", standard=Standard.PEP386)],
        "quux": [
            pydist(name="quux", dependency="python3-quux", standard=Standard.PEP440)
        ],
    }
    requires = {
        "debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt": (
            "bar",
            "baz >= 1.0",
            "quux >= 1.0a1",
        ),
    }

    def test_depends_on_bar(self) -> None:
        self.assertIn("python3-bar", self.d.depends)

    def test_depends_on_baz(self) -> None:
        self.assertIn("python3-baz (>= 1.0)", self.d.depends)

    def test_depends_on_quux(self) -> None:
        self.assertIn("python3-quux (>= 1.0~a1)", self.d.depends)


class TestRequiresCompatible(DependenciesTestCase):
    options = DHPythonOptions(guess_deps=True)
    pydist = {
        "bar": [pydist(name="bar", dependency="python3-bar")],
        "baz": [pydist(name="baz", dependency="python3-baz", standard=Standard.PEP386)],
        "qux": [pydist(name="qux", dependency="python3-qux", standard=Standard.PEP386)],
        "quux": [
            pydist(name="quux", dependency="python3-quux", standard=Standard.PEP386)
        ],
        "corge": [
            pydist(name="corge", dependency="python3-corge", standard=Standard.PEP440)
        ],
    }
    requires = {
        "debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt": (
            "bar",
            "baz ~= 1.1",
            "qux == 1.*",
            "quux",
            "corge == 1.0",
        ),
    }

    def test_depends_on_bar(self) -> None:
        self.assertIn("python3-bar", self.d.depends)

    def test_depends_on_baz(self) -> None:
        self.assertIn("python3-baz (>= 1.1), python3-baz (<< 2)", self.d.depends)

    def test_depends_on_qux(self) -> None:
        self.assertIn("python3-qux (>= 1.0), python3-qux (<< 2)", self.d.depends)

    def test_depends_on_corge(self) -> None:
        self.assertIn("python3-corge (>= 1.0), python3-corge (<< 1.1~)", self.d.depends)


class TestRequiresDistPython3(DependenciesTestCase):
    options = DHPythonOptions(guess_deps=True)
    pydist = {
        "bar": [pydist(name="bar", dependency="python3-bar")],
        "baz": [pydist(name="baz", dependency="python3-baz", standard=Standard.PEP386)],
        "qux": [pydist(name="qux", dependency="python3-qux", standard=Standard.PEP386)],
        "quux": [
            pydist(name="quux", dependency="python3-quux", standard=Standard.PEP386)
        ],
        "corge": [
            pydist(name="corge", dependency="python3-corge", standard=Standard.PEP440)
        ],
    }
    dist_info_metadata = {
        "debian/foo/usr/lib/python3/dist-packages/foo.dist-info/METADATA": (
            "Requires-Dist: bar",
            "Requires-Dist: baz >= 1.0",
            "Requires-Dist: qux == 1.*",
            "Requires-Dist: quux ~= 1.1",
            "Requires-Dist: corge == 1.0",
        ),
    }

    def test_depends_on_bar(self) -> None:
        self.assertIn("python3-bar", self.d.depends)

    def test_depends_on_baz(self) -> None:
        self.assertIn("python3-baz (>= 1.0)", self.d.depends)

    def test_depends_on_qux(self) -> None:
        self.assertIn("python3-qux (>= 1.0), python3-qux (<< 2)", self.d.depends)

    def test_depends_on_quux(self) -> None:
        self.assertIn("python3-quux (>= 1.1), python3-quux (<< 2)", self.d.depends)

    def test_depends_on_corge(self) -> None:
        self.assertIn("python3-corge (>= 1.0), python3-corge (<< 1.1~)", self.d.depends)


class TestEnvironmentMarkersDistInfo(DependenciesTestCase):
    options = DHPythonOptions(guess_deps=True, depends_section=["feature"])
    pydist = pep386(
        {
            "no_markers": "python3-no-markers",
            "os_posix": "python3-os-posix",
            "os_java": "python3-os-java",
            "sys_platform_linux": "python3-sys-platform-linux",
            "sys_platform_darwin": "python3-sys-platform-darwin",
            "platform_machine_x86_64": "python3-platform-machine-x86-64",
            "platform_machine_mips64": "python3-platform-machine-mips64",
            "platform_python_implementation_cpython": "python3-platform-python-implementation-cpython",
            "platform_python_implementation_jython": "python3-platform-python-implementation-jython",
            "platform_python_implementation_not_pypy": "python3-platform-python-implementation-not-pypy",
            "platform_python_implementation_pypy": "python3-platform-python-implementation-pypy",
            "platform_release_lt2": "python3-platform-release-lt2",
            "platform_release_ge2": "python3-platform-release-ge2",
            "platform_system_linux": "python3-platform-system-linux",
            "platform_system_windows": "python3-platform-system-windows",
            "platform_version_lt1": "python3-platform-version-lt1",
            "platform_version_ge1": "python3-platform-version-ge1",
            "python_version_ge3": "python3-python-version-ge3",
            "python_version_gt3": "python3-python-version-gt3",
            "python_version_lt3": "python3-python-version-lt3",
            "python_version_lt30": "python3-python-version-lt30",
            "python_version_lt38": "python3-python-version-lt38",
            "python_version_lt313": "python3-python-version-lt313",
            "python_version_le313": "python3-python-version-le313",
            "python_version_ge27": "python3-python-version-ge27",
            "python_version_ge313": "python3-python-version-ge313",
            "python_version_gt313": "python3-python-version-gt313",
            "python_version_eq313": "python3-python-version-eq313",
            "python_version_ne313": "python3-python-version-ne313",
            "python_version_aeq313": "python3-python-version-aeq313",
            "python_version_ceq313": "python3-python-version-ceq313",
            "python_version_weq313": "python3-python-version-weq313",
            "python_version_full_lt300": "python3-python-version-full-lt300",
            "python_version_full_lt3131": "python3-python-version-full-lt3131",
            "python_version_full_le3131": "python3-python-version-full-le3131",
            "python_version_full_ge3131": "python3-python-version-full-ge3131",
            "python_version_full_ge3131a1": "python3-python-version-full-ge3131a1",
            "python_version_full_ge3131b1post1": "python3-python-version-full-ge3131b1post1",
            "python_version_full_gt3131": "python3-python-version-full-gt3131",
            "python_version_full_eq3131": "python3-python-version-full-eq3131",
            "python_version_full_ne3131": "python3-python-version-full-ne3131",
            "python_version_full_aeq3131": "python3-python-version-full-aeq3131",
            "python_version_full_ceq3131": "python3-python-version-full-ceq3131",
            "python_version_full_weq313": "python3-python-version-full-weq313",
            "implementation_name_cpython": "python3-implementation-name-cpython",
            "implementation_name_pypy": "python3-implementation-name-pypy",
            "implementation_version_lt313": "python3-implementation-version-lt313",
            "implementation_version_ge313": "python3-implementation-version-ge313",
            "invalid_marker": "python3-invalid-marker",
            "extra_feature": "python3-extra-feature",
            "extra_test": "python3-extra-test",
            "complex_marker": "python3-complex-marker",
            "complex_marker_2": "python3-complex-marker-2",
            "no_markers_2": "python3-no-markers-2",
        }
    )
    dist_info_metadata = {
        "debian/foo/usr/lib/python3/dist-packages/foo.dist-info/METADATA": (
            "Requires-Dist: no_markers",
            "Requires-Dist: os_posix; (os_name == 'posix')",
            'Requires-Dist: os_java; os_name == "java"',
            "Requires-Dist: sys_platform_linux ; sys_platform == 'linux'",
            "Requires-Dist: sys_platform_darwin;sys_platform == 'darwin'",
            "Requires-Dist: platform_machine_x86_64; " "platform_machine == 'x86_64'",
            "Requires-Dist: platform_machine_mips64; " "platform_machine == 'mips64'",
            "Requires-Dist: platform_python_implementation_cpython; "
            "platform_python_implementation == 'CPython'",
            "Requires-Dist: platform_python_implementation_jython; "
            "platform_python_implementation == 'Jython'",
            "Requires-Dist: platform_python_implementation_pypy; "
            "platform_python_implementation == 'PyPy'",
            "Requires-Dist: platform_python_implementation_not_pypy; "
            "platform_python_implementation != 'PyPy'",
            "Requires-Dist: platform_release_lt2; platform_release < '2.0'",
            "Requires-Dist: platform_release_ge2; platform_release >= '2.0'",
            "Requires-Dist: platform_system_linux; platform_system == 'Linux'",
            "Requires-Dist: platform_system_windows; " "platform_system == 'Windows'",
            "Requires-Dist: platform_version_lt1; platform_version < '1'",
            "Requires-Dist: platform_version_ge1; platform_version >= '1'",
            "Requires-Dist: python_version_ge3; python_version >= '3'",
            "Requires-Dist: python_version_gt3; python_version > '3'",
            "Requires-Dist: python_version_lt3; python_version < '3'",
            "Requires-Dist: python_version_lt30; python_version < '3.0'",
            "Requires-Dist: python_version_lt38; python_version < '3.8'",
            "Requires-Dist: python_version_lt313; python_version < '3.13'",
            "Requires-Dist: python_version_le313; python_version <= '3.13'",
            "Requires-Dist: python_version_gt313; python_version > '3.13'",
            "Requires-Dist: python_version_ge27; python_version >= '2.7'",
            "Requires-Dist: python_version_ge313; python_version >= '3.13'",
            "Requires-Dist: python_version_eq313; python_version == '3.13'",
            "Requires-Dist: python_version_ne313; python_version != '3.13'",
            "Requires-Dist: python_version_aeq313; python_version === '3.13'",
            "Requires-Dist: python_version_ceq313; python_version ~= '3.13'",
            "Requires-Dist: python_version_weq313; python_version == '3.13.*'",
            "Requires-Dist: python_version_full_lt300; "
            "python_full_version < '3.0.0'",
            "Requires-Dist: python_version_full_lt3131; "
            "python_full_version < '3.13.1'",
            "Requires-Dist: python_version_full_le3131; "
            "python_full_version <= '3.13.1'",
            "Requires-Dist: python_version_full_gt3131; "
            "python_full_version > '3.13.1'",
            "Requires-Dist: python_version_full_ge3131; "
            "python_full_version >= '3.13.1'",
            "Requires-Dist: python_version_full_ge3131a1; "
            "python_full_version >= '3.13.1a1'",
            "Requires-Dist: python_version_full_ge3131b1post1; "
            "python_full_version >= '3.13.1b1.post1'",
            "Requires-Dist: python_version_full_eq3131; "
            "python_full_version == '3.13.1'",
            "Requires-Dist: python_version_full_ne3131; "
            "python_full_version != '3.13.1'",
            "Requires-Dist: python_version_full_aeq3131; "
            "python_full_version === '3.13.1'",
            "Requires-Dist: python_version_full_ceq3131; "
            "python_full_version ~= '3.13.1'",
            "Requires-Dist: python_version_full_weq313; "
            "python_full_version == '3.13.*'",
            "Requires-Dist: implementation_name_cpython; "
            "implementation_name == 'cpython'",
            "Requires-Dist: implementation_name_pypy; " "implementation_name == 'pypy'",
            "Requires-Dist: implementation_version_lt313; "
            "implementation_version < '3.13'",
            "Requires-Dist: implementation_version_ge313; "
            "implementation_version >= '3.13'",
            "Requires-Dist: invalid_marker; invalid_marker > '1'",
            "Requires-Dist: extra_feature; extra == 'feature'",
            "Requires-Dist: extra_test; extra == 'test'",
            "Requires-Dist: complex_marker; os_name != 'windows' "
            "and implementation_name == 'cpython'",
            'Requires-Dist: complex_marker_2; (python_version > "3.4") '
            "and extra == 'test'",
            "Requires-Dist: no_markers_2",
        ),
    }

    def test_depends_on_unmarked_packages(self) -> None:
        self.assertIn("python3-no-markers", self.d.depends)

    def test_depends_on_posix_packages(self) -> None:
        self.assertIn("python3-os-posix", self.d.depends)

    def test_skips_non_posix_packages(self) -> None:
        self.assertNotInDepends("python3-os-java")

    def test_depends_on_linux_packages(self) -> None:
        self.assertIn("python3-sys-platform-linux", self.d.depends)

    def test_skips_darwin_packages(self) -> None:
        self.assertNotInDepends("python3-sys-platform-darwin")

    def test_depends_on_x86_64_packages_on_x86_64(self) -> None:
        if platform.machine() == "x86_64":
            self.assertIn("python3-platform-machine-x86-64", self.d.depends)
        else:
            self.assertNotInDepends("python3-platform-machine-x86-64")

    def test_depends_on_mips64_packages_on_mips64(self) -> None:
        if platform.machine() == "mips64":
            self.assertIn("python3-platform-machine-mips64", self.d.depends)
        else:
            self.assertNotInDepends("python3-platform-machine-mips64")

    def test_depends_on_plat_cpython_packages(self) -> None:
        self.assertIn("python3-platform-python-implementation-cpython", self.d.depends)

    def test_depends_on_plat_pypy_packages(self) -> None:
        self.assertIn("python3-platform-python-implementation-pypy", self.d.depends)

    def test_depends_on_plat_not_pypy_packages(self) -> None:
        self.assertIn("python3-platform-python-implementation-not-pypy", self.d.depends)

    def test_skips_plat_jython_packages(self) -> None:
        self.assertNotInDepends("python3-platform-python-implementation-jython")

    def test_skips_release_lt_2_packages(self) -> None:
        self.assertNotInDepends("python3-platform-release-lt2")

    def test_skips_release_gt_2_packages(self) -> None:
        self.assertNotInDepends("python3-platform-release-ge2")

    def test_depends_on_platform_linux_packages(self) -> None:
        self.assertIn("python3-platform-system-linux", self.d.depends)

    def test_skips_platform_windows_packages(self) -> None:
        self.assertNotInDepends("python3-platform-system-windows")

    def test_skips_platfrom_version_lt_1_packages(self) -> None:
        self.assertNotInDepends("python3-platform-version-lt1")

    def test_skips_platform_version_ge_1_packages(self) -> None:
        self.assertNotInDepends("python3-platform-version-ge1")

    def test_skips_py_version_lt_3_packages(self) -> None:
        self.assertNotInDepends("python3-python-version-lt3")

    def test_elides_py_version_ge_3(self) -> None:
        self.assertIn("python3-python-version-ge3", self.d.depends)

    def test_elides_py_version_gt_3(self) -> None:
        self.assertIn("python3-python-version-gt3", self.d.depends)

    def test_skips_py_version_lt_30_packages(self) -> None:
        self.assertNotInDepends("python3-python-version-lt30")

    def test_skips_py_version_lt_38_packages(self) -> None:
        self.assertNotInDepends("python3-python-version-lt38")

    def test_depends_on_py_version_lt_313_packages(self) -> None:
        self.assertIn(
            "python3-supported-min (>= 3.13) " "| python3-python-version-lt313",
            self.d.depends,
        )

    def test_depends_on_py_version_le_313_packages(self) -> None:
        self.assertIn(
            "python3-supported-min (>> 3.13) " "| python3-python-version-le313",
            self.d.depends,
        )

    def test_depends_on_py_version_ge_27_packages(self) -> None:
        self.assertIn("python3-python-version-ge27", self.d.depends)

    def test_depends_on_py_version_ge_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-ge313 " "| python3-supported-max (<< 3.13)",
            self.d.depends,
        )

    def test_depends_on_py_version_gt_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-gt313 " "| python3-supported-max (<= 3.13)",
            self.d.depends,
        )

    def test_depends_on_py_version_eq_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-eq313 | python3-supported-max (<< 3.13) "
            "| python3-supported-min (>= 3.14)",
            self.d.depends,
        )

    def test_depends_on_py_version_ne_313_packages(self) -> None:
        # Can't be represented in Debian depends
        self.assertIn("python3-python-version-ne313", self.d.depends)

    def test_depends_on_py_version_aeq_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-aeq313 | python3-supported-max (<< 3.13) "
            "| python3-supported-min (>> 3.13)",
            self.d.depends,
        )

    def test_depends_on_py_version_ceq_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-ceq313 | python3-supported-max (<< 3.13) "
            "| python3-supported-min (>= 3.14)",
            self.d.depends,
        )

    def test_depends_on_py_version_weq_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-weq313 | python3-supported-max (<< 3.13) "
            "| python3-supported-min (>= 3.14)",
            self.d.depends,
        )

    def test_skips_py_version_full_lt_300_packages(self) -> None:
        self.assertNotInDepends("python3-python-version-full-lt300")

    def test_depends_on_py_version_full_lt_3131_packages(self) -> None:
        self.assertIn(
            "python3-supported-min (>= 3.13.1) " "| python3-python-version-full-lt3131",
            self.d.depends,
        )

    def test_depends_on_py_version_full_le_3131_packages(self) -> None:
        self.assertIn(
            "python3-supported-min (>> 3.13.1) " "| python3-python-version-full-le3131",
            self.d.depends,
        )

    def test_depends_on_py_version_full_ge_3131_packages(self) -> None:
        self.assertIn(
            "python3-python-version-full-ge3131 " "| python3-supported-max (<< 3.13.1)",
            self.d.depends,
        )

    def test_depends_on_py_version_full_ge_3131a1_packages(self) -> None:
        # With full PEP-440 parsing this should be (<< 3.13.1~a1)
        self.assertIn(
            "python3-python-version-full-ge3131a1 "
            "| python3-supported-max (<< 3.13.0)",
            self.d.depends,
        )

    def test_depends_on_py_version_full_ge_3131b1post1_packages(self) -> None:
        # With full PEP-440 parsing this should be (<< 3.13.1~b1.post1)
        self.assertIn(
            "python3-python-version-full-ge3131a1 "
            "| python3-supported-max (<< 3.13.0)",
            self.d.depends,
        )

    def test_depends_on_py_version_full_gt_3131_packages(self) -> None:
        self.assertIn(
            "python3-python-version-full-gt3131 " "| python3-supported-max (<= 3.13.1)",
            self.d.depends,
        )

    def test_depends_on_py_version_full_eq_3131_packages(self) -> None:
        self.assertIn(
            "python3-python-version-full-eq3131 "
            "| python3-supported-max (<< 3.13.1) "
            "| python3-supported-min (>> 3.13.1)",
            self.d.depends,
        )

    def test_depends_on_py_version_full_ne_3131_packages(self) -> None:
        # Can't be represented in Debian depends
        self.assertIn("python3-python-version-full-ne3131", self.d.depends)

    def test_skips_py_version_full_aeq_3131_packages(self) -> None:
        # Can't be represented in Debian depends
        self.assertNotInDepends("python3-python-version-full-aeq3131")

    def test_depends_on_py_version_full_ceq_3131_packages(self) -> None:
        self.assertIn(
            "python3-python-version-full-ceq3131 "
            "| python3-supported-max (<< 3.13.1) "
            "| python3-supported-min (>= 3.14)",
            self.d.depends,
        )

    def test_depends_on_py_version_full_weq_313_packages(self) -> None:
        self.assertIn(
            "python3-python-version-full-weq313 "
            "| python3-supported-max (<< 3.13) "
            "| python3-supported-min (>= 3.14)",
            self.d.depends,
        )

    def test_depends_on_sys_cpython_packages(self) -> None:
        self.assertIn("python3-implementation-name-cpython", self.d.depends)

    def test_depends_on_sys_pypy_packages(self) -> None:
        self.assertIn("python3-implementation-name-pypy", self.d.depends)

    def test_depends_on_sys_implementation_lt313_packages(self) -> None:
        self.assertIn(
            "python3-supported-min (>= 3.13) " "| python3-implementation-version-lt313",
            self.d.depends,
        )

    def test_depends_on_sys_implementation_ge313_packages(self) -> None:
        self.assertIn(
            "python3-implementation-version-ge313 " "| python3-supported-max (<< 3.13)",
            self.d.depends,
        )

    def test_ignores_invalid_marker(self) -> None:
        self.assertNotInDepends("python3-invalid-marker")

    def test_depends_on_extra_feature_packages(self) -> None:
        self.assertIn("python3-extra-feature", self.d.depends)

    def test_skips_extra_test_packages(self) -> None:
        self.assertNotInDepends("python3-extra-test")

    def test_ignores_complex_environment_markers(self) -> None:
        self.assertNotInDepends("python3-complex-marker")
        self.assertNotInDepends("python3-complex-marker-2")

    def test_depends_on_un_marked_dependency_after_extra(self) -> None:
        self.assertIn("python3-no-markers-2", self.d.depends)


class TestEnvironmentMarkersEggInfo(TestEnvironmentMarkersDistInfo):
    dist_info_metadata = {}
    requires = {
        "debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt": (
            "no_markers",
            "[:(os_name == 'posix')]",
            "os_posix",
            '[:os_name == "java"]',
            "os_java",
            "[:sys_platform == 'linux']",
            "sys_platform_linux",
            "[:sys_platform == 'darwin']",
            "sys_platform_darwin",
            "[:platform_machine == 'x86_64']",
            "platform_machine_x86_64",
            "[:platform_machine == 'mips64']",
            "platform_machine_mips64",
            "[:platform_python_implementation == 'CPython']",
            "platform_python_implementation_cpython",
            "[:platform_python_implementation == 'Jython']",
            "platform_python_implementation_jython",
            "[:platform_python_implementation == 'PyPy']",
            "platform_python_implementation_pypy",
            "[:platform_python_implementation != 'PyPy']",
            "platform_python_implementation_not_pypy",
            "[:platform_release < '2.0']",
            "platform_release_lt2",
            "[:platform_release >= '2.0']",
            "platform_release_ge2",
            "[:platform_system == 'Linux']",
            "platform_system_linux",
            "[:platform_system == 'Windows']",
            "platform_system_windows",
            "[:platform_version < '1']",
            "platform_version_lt1",
            "[:platform_version >= '1']",
            "platform_version_ge1",
            "[:python_version >= '3']",
            "python_version_ge3",
            "[:python_version > '3']",
            "python_version_gt3",
            "[:python_version < '3']",
            "python_version_lt3",
            "[:python_version < '3.0']",
            "python_version_lt30",
            "[:python_version < '3.8']",
            "python_version_lt313",
            "[:python_version < '3.13']",
            "python_version_lt313",
            "[:python_version <= '3.13']",
            "python_version_le313",
            "[:python_version > '3.13']",
            "python_version_gt313",
            "[:python_version >= '2.7']",
            "python_version_ge27",
            "[:python_version >= '3.13']",
            "python_version_ge313",
            "[:python_version == '3.13']",
            "python_version_eq313",
            "[:python_version != '3.13']",
            "python_version_ne313",
            "[:python_version === '3.13']",
            "python_version_aeq313",
            "[:python_version ~= '3.13']",
            "python_version_ceq313",
            "[:python_version == '3.13.*']",
            "python_version_weq313",
            "[:python_full_version < '3.0.0']",
            "python_version_full_lt300",
            "[:python_full_version < '3.13.1']",
            "python_version_full_lt3131",
            "[:python_full_version <= '3.13.1']",
            "python_version_full_le3131",
            "[:python_full_version > '3.13.1']",
            "python_version_full_gt3131",
            "[:python_full_version >= '3.13.1']",
            "python_version_full_ge3131",
            "[:python_full_version >= '3.13.1a1']",
            "python_version_full_ge3131a1",
            "[:python_full_version >= '3.13.1b1.post1']",
            "python_version_full_ge3131b1post1",
            "[:python_full_version == '3.13.1']",
            "python_version_full_eq3131",
            "[:python_full_version != '3.13.1']",
            "python_version_full_ne3131",
            "[:python_full_version === '3.13.1']",
            "python_version_full_aeq3131",
            "[:python_full_version ~= '3.13.1']",
            "python_version_full_ceq3131",
            "[:python_full_version == '3.13.*']",
            "python_version_full_weq313",
            "[:implementation_name == 'cpython']",
            "implementation_name_cpython",
            "[:implementation_name == 'pypy']",
            "implementation_name_pypy",
            "[:implementation_version < '3.13']",
            "implementation_version_lt313",
            "[:implementation_version >= '3.13']",
            "implementation_version_ge313",
            "[:invalid_marker > '1']",
            "invalid_marker",
            "[feature]",
            "extra_feature",
            "[test]",
            "extra_test",
            "[:os_name != 'windows' and implementation_name == 'cpython']",
            "complex_marker",
            "[test:(os_name != 'windows')]",
            "complex_marker_2",
        ),
    }

    def test_depends_on_un_marked_dependency_after_extra(self) -> None:
        raise unittest.SkipTest("Not possible in requires.txt")


class TestIgnoresUnusedModulesDistInfo(DependenciesTestCase):
    options = DHPythonOptions(guess_deps=True, depends_section=["feature"])
    dist_info_metadata = {
        "debian/foo/usr/lib/python3/dist-packages/foo.dist-info/METADATA": (
            "Requires-Dist: unusued-complex-module ; "
            "(sys_platform == \"darwin\") and extra == 'nativelib'",
            'Requires-Dist: unused-win-module ; (sys_platform == "win32")',
            "Requires-Dist: unused-extra-module ; extra == 'unused'",
        ),
    }
    parse = False

    def test_ignores_unused_dependencies(self) -> None:
        if not hasattr(self, "assertLogs"):
            raise unittest.SkipTest("Requires Python >= 3.4")
        with self.assertLogs(logger="dhpython", level=logging.INFO) as logs:
            self.d.parse(self.prepared_stats, self.options)
        for line in logs.output:
            self.assertTrue(
                line.startswith("INFO:dhpython:Ignoring complex environment marker"),
                "Expecting only complex environment marker messages, but "
                "got: {}".format(line),
            )


class TestIgnoresUnusedModulesEggInfo(DependenciesTestCase):
    options = DHPythonOptions(guess_deps=True, depends_section=["feature"])
    requires = {
        "debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt": (
            "[nativelib:(sys_platform == 'darwin')]",
            "unusued-complex-module",
            "[:sys_platform == 'win32']",
            "unused-win-module",
            "[unused]",
            "unused-extra-module",
        )
    }
    parse = False

    def test_ignores_unused_dependencies(self) -> None:
        if not hasattr(self, "assertNoLogs"):
            raise unittest.SkipTest("Requires Python >= 3.10")
        with self.assertNoLogs(logger="dhpython", level=logging.INFO):
            self.d.parse(self.prepared_stats, self.options)


class TestCExtensionPython3(DependenciesTestCase):
    stats = {
        "public_vers": {Version("3")},
        "ext_vers": {Version("3.14")},
    }

    def test_python_depends(self) -> None:
        self.assertIn("python3 (>= 3.14~)", self.d.depends)
        self.assertIn("python3 (<< 3.15)", self.d.depends)


class TestCExtensionsPython3(DependenciesTestCase):
    stats = {
        "public_vers": {Version("3")},
        "ext_vers": {Version("3.14"), Version("3.13")},
    }

    def test_python_depends(self) -> None:
        self.assertIn("python3 (>= 3.13~)", self.d.depends)
        self.assertIn("python3 (<< 3.15)", self.d.depends)


class TestStableABIExtensionPython3(DependenciesTestCase):
    stats = {
        "public_vers": {Version("3")},
        "ext_stableabi": {"foo.abi3.so"},
    }

    def test_python_depends(self) -> None:
        self.assertIn("python3", self.d.depends)


class TestMixedExtensionsPython3(DependenciesTestCase):
    stats = {
        "public_vers": {Version("3")},
        "ext_vers": {Version("3.14"), Version("3.13")},
        "ext_stableabi": {"foo.abi3.so"},
    }

    def test_python_depends(self) -> None:
        self.assertIn("python3 (>= 3.13~)", self.d.depends)
        self.assertIn("python3 (<< 3.15)", self.d.depends)


class TestUnversionedExtensionPython3(DependenciesTestCase):
    stats = {
        "public_vers": {Version("3")},
        "ext_no_version": {"foo.so"},
    }

    def test_python_depends(self) -> None:
        # This test is to maintain the status-quo where this case is only
        # handled in private modules. That's *probably* OK?
        self.assertNotInDepends("python3")



# Copyright © 2013-2019 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import difflib
import logging
import os
import re
import sys
from collections.abc import Iterable, Sequence
from filecmp import cmp as cmpfile
from os.path import lexists, exists, isdir, islink, join, realpath, split, splitext
from pathlib import Path
from shutil import rmtree
from stat import ST_MODE, S_IXUSR, S_IXGRP, S_IXOTH
from typing import Literal, TypedDict

from dhpython import CACHE_DIRECTORIES, MULTIARCH_DIR_TPL
from dhpython.tools import fix_shebang, clean_egg_name
from dhpython.interpreter import EXTFILE_RE, Interpreter
from dhpython.options import DHPythonOptions
from dhpython.version import Version

log = logging.getLogger("dhpython")


def fix_locations(
    package: str,
    interpreter: Interpreter,
    versions: Sequence[Version],
    options: DHPythonOptions,
) -> None:
    """Move files to the right location."""
    # make a copy since we change version later
    interpreter = Interpreter(interpreter)

    for version in versions:
        interpreter.version = version

        dstdir = interpreter.sitedir(package)
        for srcdir in interpreter.old_sitedirs(package):
            if isdir(srcdir):
                # TODO: what about relative symlinks?
                log.debug("moving files from %s to %s", srcdir, dstdir)
                share_files(srcdir, dstdir, interpreter, options)
                try:
                    os.removedirs(srcdir)
                except OSError:
                    pass

        # do the same with debug locations
        dstdir = interpreter.sitedir(package, gdb=True)
        for srcdir in interpreter.old_sitedirs(package, gdb=True):
            if isdir(srcdir):
                log.debug("moving files from %s to %s", srcdir, dstdir)
                share_files(srcdir, dstdir, interpreter, options)
                try:
                    os.removedirs(srcdir)
                except OSError:
                    pass


def share_files(
    srcdir: str,
    dstdir: str,
    interpreter: Interpreter,
    options: DHPythonOptions,
) -> None:
    """Try to move as many files from srcdir to dstdir as possible."""
    for i in os.listdir(srcdir):
        fpath1 = join(srcdir, i)
        if not lexists(fpath1):  # removed in rename_ext
            continue
        if i.endswith(".pyc"):  # f.e. when tests were invoked on installed files
            os.remove(fpath1)
            continue
        if i in CACHE_DIRECTORIES:
            if os.path.isdir(fpath1):
                rmtree(fpath1)
            else:
                os.remove(fpath1)
            continue
        if not options.no_ext_rename and splitext(i)[-1] == ".so":
            # try to rename extension here as well (in :meth:`scan` info about
            # Python version is gone)
            version = interpreter.parse_public_dir(srcdir)
            if version and version is not True:
                fpath1 = Scan.rename_ext(fpath1, interpreter, version)
                i = split(fpath1)[-1]
        if srcdir.endswith(".dist-info"):
            if i.startswith(
                ("LICENCE", "LICENSE", "COPYING", "NOTICE", "AUTHORS")
            ) or i in ("licenses", "license_files"):
                if isdir(fpath1):
                    rmtree(fpath1)
                else:
                    os.remove(fpath1)
                continue
        fpath2 = join(dstdir, i)
        if not isdir(fpath1) and not exists(fpath2):
            # do not rename directories here - all .so files have to be renamed first
            os.renames(fpath1, fpath2)
            continue
        ext_m = EXTFILE_RE.search(i)
        if islink(fpath1):
            # move symlinks without changing them if they point to the same place
            if not exists(fpath2):
                os.renames(fpath1, fpath2)
            elif realpath(fpath1) == realpath(fpath2):
                os.remove(fpath1)
        elif isdir(fpath1):
            share_files(fpath1, fpath2, interpreter, options)
        elif cmpfile(fpath1, fpath2, shallow=False):
            os.remove(fpath1)
        elif (
            ext_m and ext_m.group("stableabi") and interpreter.parse_public_dir(srcdir)
        ):
            log.warning("%s differs from previous one, removing anyway (%s)", i, srcdir)
            os.remove(fpath1)
        elif srcdir.endswith(".dist-info"):
            # dist-info file that differs... try merging
            if i == "WHEEL":
                os.remove(fpath1)
            elif i == "RECORD":
                os.remove(fpath1)
            else:
                log.warning("No merge driver for dist-info file %s", i)
        else:
            # The files differed so we cannot collapse them.
            log.warning("Paths differ: %s and %s", fpath1, fpath2)
            if options.verbose and not i.endswith((".so", ".a")):
                with open(fpath1, encoding="UTF-8") as fp1:
                    fromlines = fp1.readlines()
                with open(fpath2, encoding="UTF-8") as fp2:
                    tolines = fp2.readlines()
                diff = difflib.unified_diff(fromlines, tolines, fpath1, fpath2)
                sys.stderr.writelines(diff)

    try:
        os.removedirs(srcdir)
    except OSError:
        pass


## Functions to merge parts of the .dist-info metadata directory together


def missing_lines(src: str | Path, dst: str | Path) -> list[str]:
    """Find all the lines in the text file src that are not in dst"""
    with open(dst, encoding="UTF-8") as fh:
        current = {k: None for k in fh.readlines()}

    missing = []
    with open(src, encoding="UTF-8") as fh:
        for line in fh.readlines():
            if line not in current:
                missing.append(line)

    return missing


def merge_WHEEL(src: str | Path, dst: str | Path) -> int:
    """Merge the source .dist-info/WHEEL file into the destination

    Note that after editing the WHEEL file, the sha256 included in
    the .dist-info/RECORD file will be incorrect and will need fixing
    using the fix_merged_RECORD() function.
    """
    log.debug("Merging WHEEL file %s into %s", src, dst)
    missing = missing_lines(src, dst)
    with open(dst, "a", encoding="UTF-8") as fh:
        for line in missing:
            if line.startswith("Tag: "):
                fh.write(line)
            else:
                log.warning("WHEEL merge discarded line %s", line)

    return len(missing)


def write_INSTALLER(distdir: str) -> None:
    """Write 'debian' as the INSTALLER"""
    log.debug("Writing INSTALLER in %s", distdir)
    installer = join(distdir, "INSTALLER")
    with open(installer, "w", encoding="UTF-8") as f:
        f.write("debian\n")


def remove_RECORD(distdir: str) -> None:
    """Remove RECORD from a .dist-info.

    Having this breaks multi-arch installation, as the contents differ.
    """
    log.debug("Deleting RECORD in %s", distdir)
    try:
        os.remove(join(distdir, "RECORD"))
    except FileNotFoundError:
        pass


ScanResult = TypedDict(
    "ScanResult",
    {
        "requires.txt": set[str],
        "egg-info": set[str],
        "dist-info": set[str],
        "nsp.txt": set[str],
        "shebangs": set[Interpreter],
        "public_vers": set[Version],
        "private_dirs": dict[str, "ScanResult"],
        "compile": bool,
        "ext_vers": set[Version],
        "ext_stableabi": set[str],
        "ext_no_version": set[str],
    },
)


class Scan:
    UNWANTED_DIRS = re.compile(r".*/__pycache__(/.*)?$")
    UNWANTED_FILES = re.compile(r".*\.py[co]$")

    interprter: Interpreter
    impl: str
    package: str
    proot: str
    dpath: str | None | Literal[False]
    options: DHPythonOptions
    result: ScanResult

    def __init__(
        self,
        interpreter: Interpreter,
        package: str,
        dpath: str | None = None,
        *,
        options: DHPythonOptions,
    ) -> None:
        self.interpreter = interpreter
        self.impl = interpreter.impl

        self.package = package

        if not dpath:
            self.proot = "debian/%s" % self.package
        else:
            dpath = dpath.strip("/")
            self.proot = join("debian", self.package, dpath)
        self.dpath = dpath
        del dpath

        self.options = options
        self.result = {
            "requires.txt": set(),
            "egg-info": set(),
            "dist-info": set(),
            "nsp.txt": set(),
            "shebangs": set(),
            "public_vers": set(),
            "private_dirs": {},
            "compile": False,
            "ext_vers": set(),
            "ext_stableabi": set(),
            "ext_no_version": set(),
        }

        for root, dirs, file_names in os.walk(self.proot):
            if interpreter.should_ignore(root):
                del dirs[:]
                continue

            self.current_private_dir = self.current_pub_version = None
            version = interpreter.parse_public_dir(root)
            if version:
                self.current_dir_is_public = True
                if version is True:
                    version = None
                else:
                    self.current_pub_version = version
            else:
                self.current_dir_is_public = False

            if self.current_dir_is_public:
                if root.endswith("-packages"):
                    if version is not None:
                        assert isinstance(version, Version)
                        self.result["public_vers"].add(version)
                    for name in dirs[:]:
                        if name in ("test", "tests") or name.startswith("."):
                            log.debug("removing dist-packages/%s", name)
                            rmtree(join(root, name))
                            dirs.remove(name)
                    for name in file_names[:]:
                        if name.startswith("."):
                            log.debug("removing dist-packages/%s", name)
                            os.remove(join(root, name))
                            file_names.remove(name)
            else:
                self.current_private_dir = self.check_private_dir(root)
                if not self.current_private_dir:
                    # i.e. not a public dir and not a private dir
                    if self.is_bin_dir(root):
                        self.handle_bin_dir(root, file_names)
                    else:  # not a public, private or bin directory
                        # continue with a subdirectory
                        continue

            for name in dirs[:]:
                dpath = join(root, name)
                if self.is_unwanted_dir(dpath):
                    rmtree(dpath)
                    dirs.remove(name)

            if self.is_dist_dir(root):
                self.handle_dist_dir(root, file_names)
                continue

            if self.is_egg_dir(root):
                self.handle_egg_dir(root, file_names)
                continue

            # check files
            for fn in sorted(file_names):
                # sorted() to make sure .so files are handled before .so.foo
                fpath = join(root, fn)

                if self.is_unwanted_file(fpath):
                    log.debug("removing unwanted: %s", fpath)
                    os.remove(fpath)
                    continue

                if self.is_egg_file(fpath):
                    self.handle_egg_file(fpath)
                    continue

                if not exists(fpath):
                    # possibly removed while handling .so symlinks
                    if islink(fpath) and ".so." in split(fpath)[-1]:
                        # dangling symlink to (now removed/renamed) .so file
                        # which wasn't removed yet (see test203's quux.so.0)
                        log.info("removing dangling symlink: %s", fpath)
                        os.remove(fpath)
                    continue

                fext = splitext(fn)[-1][1:]
                ext_m = EXTFILE_RE.search(fn)
                if ext_m:
                    if not self.options.no_ext_rename:
                        fpath = self.rename_ext(fpath, interpreter, version)
                        ext_m = EXTFILE_RE.search(os.path.basename(fpath))
                        assert ext_m is not None
                    ver = version

                    if ext_m["ver"]:
                        assert ext_m["ver"].startswith("3")
                        if ext_m["ver"] == "3":
                            ver = None
                        ver = Version("3." + ext_m["ver"][1:])
                    else:
                        ver = None
                    if ver:
                        self.current_result.setdefault("ext_vers", set()).add(ver)
                    elif ext_m["stableabi"]:
                        self.current_result.setdefault("ext_stableabi", set()).add(
                            fpath
                        )
                    else:
                        self.current_result.setdefault("ext_no_version", set()).add(
                            fpath
                        )

                if self.current_private_dir:
                    if exists(fpath) and fext != "so":
                        mode = os.stat(fpath)[ST_MODE]
                        if mode & S_IXUSR or mode & S_IXGRP or mode & S_IXOTH:
                            if (
                                options.no_shebang_rewrite
                                or fix_shebang(fpath, self.options.shebang)
                            ) and not self.options.ignore_shebangs:
                                try:
                                    res = Interpreter.from_file(fpath)
                                except Exception as e:
                                    log.debug("cannot parse shebang %s: %s", fpath, e)
                                else:
                                    self.current_result.setdefault(
                                        "shebangs", set()
                                    ).add(res)

                if fext == "py":
                    self.current_result["compile"] = True

            if not dirs and not self.current_private_dir:
                try:
                    os.removedirs(root)
                except OSError:
                    pass

        log.debug("package %s details = %s", package, self.result)

    @property
    def current_result(self) -> ScanResult:
        if self.current_private_dir:
            return self.result["private_dirs"].setdefault(
                self.current_private_dir,
                {
                    "requires.txt": set(),
                    "egg-info": set(),
                    "dist-info": set(),
                    "nsp.txt": set(),
                    "shebangs": set(),
                    "public_vers": set(),
                    "private_dirs": {},
                    "compile": False,
                    "ext_vers": set(),
                    "ext_stableabi": set(),
                    "ext_no_version": set(),
                },
            )
        return self.result

    def is_unwanted_dir(self, dpath: str) -> bool:
        return self.__class__.UNWANTED_DIRS.match(dpath) is not None

    def is_unwanted_file(self, fpath: str) -> bool:
        if self.__class__.UNWANTED_FILES.match(fpath):
            return True
        if (
            self.current_dir_is_public
            and self.is_dbg_package
            and self.options.clean_dbg_pkg
            and splitext(fpath)[-1][1:] not in ("so", "h")
        ):
            return True
        return False

    @property
    def private_dirs_to_check(self) -> list[str]:
        if self.dpath:
            # scan private directory *only*
            return [self.dpath]

        if self.dpath is False:
            result = []
        else:
            result = [
                i % self.package
                for i in (
                    "usr/lib/%s",
                    "usr/lib/games/%s",
                    "usr/share/%s",
                    "usr/share/games/%s",
                )
            ]
        return result

    @property
    def is_dbg_package(self) -> bool:
        # return self.interpreter.debug
        return self.package.endswith("-dbg")

    def check_private_dir(self, dpath: str) -> str | None:
        """Return private dir's root if it's a private dir."""
        for i in self.private_dirs_to_check:
            if dpath.startswith(join("debian", self.package, i)):
                return "/" + i
        return None

    @staticmethod
    def rename_ext(
        fpath: str, interpreter: Interpreter, current_pub_version: Version | None = None
    ) -> str:
        """Add multiarch triplet, etc. Return new name.

        This method is invoked for all .so files in public or private directories.
        """
        # current_pub_version - version parsed from dist-packages (True if unversioned)
        # i.e. if it's not None - it's a public dist-packages directory

        path, fname = fpath.rsplit("/", 1)
        if current_pub_version is not None and islink(fpath):
            # replace symlinks with extensions in dist-packages directory
            dstfpath = fpath
            links = set()
            while islink(dstfpath):
                links.add(dstfpath)
                dstfpath = join(path, os.readlink(dstfpath))
            if exists(dstfpath) and ".so." in split(dstfpath)[-1]:
                # rename .so.$FOO symlinks, remove other ones
                for lpath in links:
                    log.info("removing symlink: %s", lpath)
                    os.remove(lpath)
                log.info("renaming %s to %s", dstfpath, fname)
                os.rename(dstfpath, fpath)

        if MULTIARCH_DIR_TPL.match(fpath):
            # ignore /lib/i386-linux-gnu/, /usr/lib/x86_64-kfreebsd-gnu/, etc.
            return fpath

        new_fn = interpreter.check_extname(fname, current_pub_version)
        if new_fn:
            # TODO: what about symlinks pointing to this file
            new_fpath = join(path, new_fn)
            if exists(new_fpath):
                log.warning(
                    "destination file exist, cannot rename %s to %s", fname, new_fn
                )
            else:
                log.info("renaming %s to %s", fname, new_fn)
                os.rename(fpath, new_fpath)
            return new_fpath
        return fpath

    def is_bin_dir(self, dpath: str) -> bool:
        """Check if dir is one from PATH ones."""
        # dname = debian/packagename/usr/games
        spath = dpath.strip("/").split("/", 4)
        if len(spath) > 4:
            return False  # assume bin directories don't have subdirectories
        if dpath.endswith(("/sbin", "/bin", "/usr/games")):
            # /(s)bin or /usr/(s)bin or /usr/games
            return True
        return False

    def handle_bin_dir(self, dpath: str, file_names: Iterable[str]) -> None:
        if self.options.no_shebang_rewrite or self.options.ignore_shebangs:
            return
        for fn in file_names:
            fpath = join(dpath, fn)
            if fix_shebang(fpath, self.options.shebang):
                try:
                    res = Interpreter.from_file(fpath)
                except Exception as e:
                    log.debug("cannot parse shebang %s: %s", fpath, e)
                else:
                    self.result["shebangs"].add(res)

    def is_egg_dir(self, dname: str) -> bool:
        """Check if given directory contains egg-info."""
        return dname.endswith(".egg-info")

    def handle_egg_dir(self, dpath: str, file_names: list[str]) -> None:
        # NOTE: Mutates file_names
        path, dname = dpath.rsplit("/", 1)
        if self.is_dbg_package and self.options.clean_dbg_pkg:
            rmtree(dpath)
            return

        clean_name = clean_egg_name(dname)
        if clean_name != dname:
            if exists(join(path, clean_name)):
                log.info("removing %s (%s is already available)", dname, clean_name)
                rmtree(dpath)
                return
            else:
                log.info("renaming %s to %s", dname, clean_name)
                os.rename(dpath, join(path, clean_name))
                dname = clean_name
                dpath = join(path, dname)
        if file_names:
            if "requires.txt" in file_names:
                self.result["requires.txt"].add(join(dpath, "requires.txt"))
            if "namespace_packages.txt" in file_names:
                self.result["nsp.txt"].add(join(dpath, "namespace_packages.txt"))
            if "PKG-INFO" in file_names:
                self.result["egg-info"].add(join(dpath, "PKG-INFO"))
            if "SOURCES.txt" in file_names:
                os.remove(join(dpath, "SOURCES.txt"))
                file_names.remove("SOURCES.txt")

    def is_egg_file(self, fpath: str) -> bool:
        """Check if given file contains egg-info."""
        return fpath.endswith(".egg-info")

    def handle_egg_file(self, fpath: str) -> None:
        root, name = fpath.rsplit("/", 1)
        clean_name = clean_egg_name(name)
        if clean_name != name:
            if exists(join(root, clean_name)):
                log.info("removing %s (%s is already available)", name, clean_name)
                os.remove(fpath)
            else:
                log.info("renaming %s to %s", name, clean_name)
                os.rename(fpath, join(root, clean_name))
        self.result["egg-info"].add(join(root, clean_name))

    def is_dist_dir(self, dname: str) -> bool:
        """Check if given directory contains dist-info."""
        return dname.endswith(".dist-info")

    def handle_dist_dir(self, dpath: str, file_names: Iterable[str]) -> None:
        if self.is_dbg_package and self.options.clean_dbg_pkg:
            rmtree(dpath)
            return

        if file_names:
            write_INSTALLER(dpath)
            remove_RECORD(dpath)
            if "METADATA" in file_names:
                self.result["dist-info"].add(join(dpath, "METADATA"))

    def cleanup(self) -> None:
        if self.is_dbg_package and self.options.clean_dbg_pkg:
            # remove empty directories in -dbg packages
            proot = self.proot + "/usr/lib"
            for root, _, file_names in os.walk(proot, topdown=False):
                if "-packages/" in root and not file_names:
                    try:
                        os.removedirs(root)
                    except Exception:
                        pass



from dataclasses import fields
from unittest import TestCase

from dhpython.build.options import build_parser, PybuildOptions


class TestPybuildOptions(TestCase):
    def test_args_match_options(self) -> None:
        parser = build_parser()
        parser_options = {action.dest for action in parser._get_optional_actions()} | {
            action.dest for action in parser._get_positional_actions()
        }
        # Built-ins
        parser_options -= {"help", "version"}

        options = {field.name for field in fields(PybuildOptions)}
        # Added in parse_args()
        options -= {"custom_tests"}

        self.assertEqual(parser_options, options)



import os
from collections.abc import Sequence
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from unittest import TestCase

from dhpython.interpreter import Interpreter
from dhpython.fs import Scan, merge_WHEEL, share_files
from dhpython.options import DHPythonOptions
from dhpython.version import Version


class FSTestCase(TestCase):
    files: dict[str, Sequence[str]] = {}

    def setUp(self) -> None:
        self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tempdir.cleanup)
        temp_path = Path(self.tempdir.name)
        for fn, contents in self.files.items():
            path = temp_path / fn
            setattr(self, path.name, path)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text("\n".join(contents) + "\n")

    def assertFileContents(self, path: Path, contents: str | Sequence[str]) -> None:
        """Assert that the contents of path is contents

        Contents may be specified as a list of strings, one per line, without
        line-breaks.
        """
        if isinstance(contents, (list, tuple)):
            contents = "\n".join(contents) + "\n"
        assert isinstance(contents, str)
        with path.open("r") as f:
            self.assertMultiLineEqual(contents, f.read())


class MergeTagsTest(FSTestCase):
    a: Path
    b: Path
    files = {
        "a": ("foo", "Tag: A"),
        "b": ("foo", "Tag: B"),
    }

    def test_merge_wheel(self) -> None:
        merge_WHEEL(self.a, self.b)
        self.assertFileContents(self.b, ("foo", "Tag: B", "Tag: A"))


class ShareFilesTestCase(FSTestCase):
    impl = "cpython3"
    options: dict[str, Any] = {}

    def setUp(self) -> None:
        super().setUp()
        self.destdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.destdir.cleanup)
        self.interpreter = Interpreter(self.impl)

    def share_files(self, options: dict[str, Any] | None = None) -> None:
        if options is None:
            options = self.options
        share_files(
            self.tempdir.name,
            self.destdir.name,
            self.interpreter,
            DHPythonOptions(**options),
        )

    def destPath(self, name: str) -> Path:
        return Path(self.destdir.name) / name


class HatchlingLicenseTest(ShareFilesTestCase):
    files = {
        "foo.dist-info/license_files/LICENSE.txt": ("foo"),
        "foo.dist-info/licenses/COPYING": ("foo"),
        "foo.dist-info/RECORD": (
            "foo.dist-info/license_files/LICENSE.txt,sha256=2c26b46b68ffc68ff99"
            "b453c1d30413413422d706483bfa0f98a5e886266e7ae,4",
            "foo.dist-info/licenses/COPYING,sha256=2c26b46b68ffc68ff99b453c1d30"
            "413413422d706483bfa0f98a5e886266e7ae,4",
            "foo.dist-info/WHEEL,sha256=447fb61fa39a067229e1cce8fc0953bfced53ea"
            "c85d1844f5940f51c1fcba725,6",
        ),
        "foo.dist-info/WHEEL": ("foo"),
    }

    def test_removes_license_files(self) -> None:
        self.share_files()
        self.assertFalse(
            self.destPath("foo.dist-info/license_files/LICENSE.txt").exists()
        )
        self.assertFalse(self.destPath("foo.dist-info/licenses/COPYING").exists())


class FlitLicenseTest(ShareFilesTestCase):
    files = {
        "foo.dist-info/COPYING": ("foo"),
        "foo.dist-info/COPYING.LESSER": ("foo"),
        "foo.dist-info/RECORD": (
            "foo.dist-info/COPYING,sha256=2c26b46b68ffc68ff99b453c1d30413413422"
            "d706483bfa0f98a5e886266e7ae,4",
            "foo.dist-info/COPYING.LESSER,sha256=2c26b46b68ffc68ff99b453c1d3041"
            "3413422d706483bfa0f98a5e886266e7ae,4",
            "foo.dist-info/WHEEL,sha256=447fb61fa39a067229e1cce8fc0953bfced53ea"
            "c85d1844f5940f51c1fcba725,6",
        ),
        "foo.dist-info/WHEEL": ("foo"),
    }

    def test_removes_license_files(self) -> None:
        self.share_files()
        self.assertFalse(self.destPath("foo.dist-info/COPYING.LESSER").exists())
        self.assertFalse(self.destPath("foo.dist-info/COPYING").exists())


class CExtensionsRenameTest(ShareFilesTestCase):
    def setUp(self) -> None:
        ver = Interpreter(self.impl).default_version
        self.files = {
            f"usr/lib/python{ver}/foo.so": ("binary-data"),
        }
        self.ext_filename = f"usr/lib/python{ver}/foo.so"
        super().setUp()

    def test_python_extensions(self) -> None:
        self.share_files()
        self.assertFalse(self.destPath(self.ext_filename).exists())
        expected_name = self.interpreter.check_extname(self.ext_filename)
        self.assertTrue(expected_name)
        assert expected_name
        self.assertTrue(self.destPath(expected_name).exists())

    def test_no_ext_rename(self) -> None:
        self.share_files(options={"no_ext_rename": True})
        self.assertTrue(self.destPath(self.ext_filename).exists())


class ScanTestCase(FSTestCase):
    impl = "cpython3"
    options: dict[str, Any] = {}
    package = "python3-test"

    def setUp(self) -> None:
        super().setUp()
        pwd = os.getcwd()
        try:
            os.chdir(self.tempdir.name)
            self.scan = Scan(
                interpreter=Interpreter(self.impl),
                package=self.package,
                options=DHPythonOptions(**self.options),
            )
        finally:
            os.chdir(pwd)


class UnversionedExtensionScanTest(ScanTestCase):
    files = {
        "debian/python3-test/usr/lib/python3/dist-packages/foo.so": "contents",
    }
    options = {
        "no_ext_rename": True,
    }

    def test_scan(self) -> None:
        self.assertEqual(self.scan.result["public_vers"], {Version("3")})
        self.assertEqual(self.scan.result["ext_vers"], set())
        self.assertEqual(
            self.scan.result["ext_no_version"],
            {"debian/python3-test/usr/lib/python3/dist-packages/foo.so"},
        )
        self.assertEqual(self.scan.result["ext_stableabi"], set())


class UnversionedExtensionRenamedScanTest(ScanTestCase):
    files = {
        "debian/python3-test/usr/lib/python3/dist-packages/foo.so": "contents",
    }

    def test_scan(self) -> None:
        version = Interpreter(self.impl).default_version
        self.assertEqual(self.scan.result["public_vers"], {Version("3")})
        self.assertEqual(self.scan.result["ext_vers"], {version})
        self.assertEqual(self.scan.result["ext_no_version"], set())
        self.assertEqual(self.scan.result["ext_stableabi"], set())


class StableABIExtensionDependenciesTest(ScanTestCase):
    files = {
        (
            "debian/python3-test/usr/lib/python3/dist-packages/"
            "foo.abi3-x86_64-linux-gnu.so"
        ): "contents",
    }

    def test_scan(self) -> None:
        self.assertEqual(self.scan.result["public_vers"], {Version("3")})
        self.assertEqual(self.scan.result["ext_vers"], set())
        self.assertEqual(self.scan.result["ext_no_version"], set())
        self.assertEqual(
            self.scan.result["ext_stableabi"],
            {
                "debian/python3-test/usr/lib/python3/dist-packages/"
                "foo.abi3-x86_64-linux-gnu.so"
            },
        )


class VersionedExtensionDependenciesTest(ScanTestCase):
    files = {
        (
            "debian/python3-test/usr/lib/python3/dist-packages/"
            "foo.cpython-314-x86_64-linux-gnu.so"
        ): "contents",
    }

    def test_scan(self) -> None:
        self.assertEqual(self.scan.result["public_vers"], {Version("3")})
        self.assertEqual(self.scan.result["ext_vers"], {Version("3.14")})
        self.assertEqual(self.scan.result["ext_no_version"], set())
        self.assertEqual(self.scan.result["ext_stableabi"], set())



# Copyright © 2025 Stefano Rivera <stefanor@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from argparse import ArgumentParser, Action, Namespace
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Any, override
from os import environ, getcwd

from dhpython.version import Version

DEFAULT_EXT_PATTERN = r"\.so(\.[^/]*)?$"
DEFAULT_INSTALL_DIR = "/usr/lib/python{version}/dist-packages"

@dataclass
class PybuildOptions:
    verbose: bool = environ.get("PYBUILD_VERBOSE") == "1"
    quiet: bool = environ.get("PYBUILD_QUIET") == "1"
    really_quiet: bool = environ.get("PYBUILD_RQUIET") == "1"
    detect_only: bool = False
    clean_only: bool = False
    configure_only: bool = False
    build_only: bool = False
    install_only: bool = False
    test_only: bool = False
    autopkgtest_only: bool = False
    list_systems: bool = False
    print_args: list[str] = field(default_factory=list)
    before_clean: str | None = None
    clean_args: str | None = None
    after_clean: str | None = None
    before_configure: str | None = None
    configure_args: str | None = None
    after_configure: str | None = None
    before_build: str | None = None
    build_args: str | None = None
    after_build: str | None = None
    before_install: str | None = None
    install_args: str | None = None
    after_install: str | None = None
    before_test: str | None = None
    test_args: str | None = None
    after_test: str | None = None
    test_nose: bool = environ.get("PYBUILD_TEST_NOSE") == "1"
    test_nose2: bool = environ.get("PYBUILD_TEST_NOSE2") == "1"
    test_pytest: bool = environ.get("PYBUILD_TEST_PYTEST") == "1"
    test_tox: bool = environ.get("PYBUILD_TEST_TOX") == "1"
    test_stestr: bool = environ.get("PYBUILD_TEST_STESTR") == "1"
    test_unittest: bool = environ.get("PYBUILD_TEST_UNITTEST") == "1"
    test_custom: bool = environ.get("PYBUILD_TEST_CUSTOM") == "1"
    dir: str = environ.get("PYBUILD_DIR", getcwd())
    # We handle this default manually, because --name affects it
    destdir: str | None = None
    ext_destdir: str | None = environ.get("PYBUILD_EXT_DESTDIR")
    ext_pattern: str = environ.get("PYBUILD_EXT_PATTERN", DEFAULT_EXT_PATTERN)
    ext_sub_pattern: str | None = environ.get("PYBUILD_EXT_SUB_PATTERN")
    ext_sub_repl: str | None = environ.get("PYBUILD_EXT_SUB_REPL")
    install_dir: str = DEFAULT_INSTALL_DIR
    name: str | None = environ.get("PYBUILD_NAME")
    system: str | None = environ.get("PYBUILD_SYSTEM")
    versions: list[Version] = field(default_factory=list)
    interpreter: list[str] = field(default_factory=list)
    disable: str | None = None
    parallel: int = 1
    custom_tests: bool = False


class AppendVersions(Action):
    """
    Append possibly-space-separated versions.

    This allows the use of -p "$(py3versions -rv)".
    """

    @override
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        assert isinstance(values, str)
        for value in values.split():
            getattr(namespace, self.dest).append(Version(value))


def build_parser() -> ArgumentParser:
    usage = "%(prog)s [ACTION] [BUILD SYSTEM ARGS] [DIRECTORIES] [OPTIONS]"
    parser = ArgumentParser(usage=usage)
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="turn verbose mode on",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="doesn't show external command's output",
    )
    parser.add_argument(
        "-qq",
        "--really-quiet",
        action="store_true",
        help="be quiet",
    )
    parser.add_argument("--version", action="version", version="%(prog)s DEVELV")

    action = parser.add_argument_group(
        "ACTION",
        """The default is to build,
        install and test the library using detected build system version by
        version. Selecting one of following actions, will invoke given action
        for all versions - one by one - which (contrary to the default action)
        in some build systems can overwrite previous results.""",
    )
    action.add_argument(
        "--detect",
        action="store_true",
        dest="detect_only",
        help="return the name of detected build system",
    )
    action.add_argument(
        "--clean",
        action="store_true",
        dest="clean_only",
        help="clean files using auto-detected build system specific methods",
    )
    action.add_argument(
        "--configure",
        action="store_true",
        dest="configure_only",
        help="invoke configure step for all requested Python versions",
    )
    action.add_argument(
        "--build",
        action="store_true",
        dest="build_only",
        help="invoke build step for all requested Python versions",
    )
    action.add_argument(
        "--install",
        action="store_true",
        dest="install_only",
        help="invoke install step for all requested Python versions",
    )
    action.add_argument(
        "--test",
        action="store_true",
        dest="test_only",
        help="invoke tests for auto-detected build system",
    )
    action.add_argument(
        "--autopkgtest",
        action="store_true",
        dest="autopkgtest_only",
        help="invoke autopkgtests for auto-detected build system",
    )
    action.add_argument(
        "--list-systems",
        action="store_true",
        help="list available build systems and exit",
    )
    action.add_argument(
        "--print",
        action="append",
        dest="print_args",
        help="print pybuild's internal parameters",
    )

    arguments = parser.add_argument_group(
        "BUILD SYSTEM ARGS",
        """
        Additional arguments passed to the build system.
        --system=custom requires complete command.""",
    )
    arguments.add_argument(
        "--before-clean", metavar="CMD", help="invoked before the clean command"
    )
    arguments.add_argument("--clean-args", metavar="ARGS")
    arguments.add_argument(
        "--after-clean", metavar="CMD", help="invoked after the clean command"
    )

    arguments.add_argument(
        "--before-configure", metavar="CMD", help="invoked before the configure command"
    )
    arguments.add_argument("--configure-args", metavar="ARGS")
    arguments.add_argument(
        "--after-configure", metavar="CMD", help="invoked after the configure command"
    )

    arguments.add_argument(
        "--before-build", metavar="CMD", help="invoked before the build command"
    )
    arguments.add_argument("--build-args", metavar="ARGS")
    arguments.add_argument(
        "--after-build", metavar="CMD", help="invoked after the build command"
    )

    arguments.add_argument(
        "--before-install", metavar="CMD", help="invoked before the install command"
    )
    arguments.add_argument("--install-args", metavar="ARGS")
    arguments.add_argument(
        "--after-install", metavar="CMD", help="invoked after the install command"
    )

    arguments.add_argument(
        "--before-test", metavar="CMD", help="invoked before the test command"
    )
    arguments.add_argument("--test-args", metavar="ARGS")
    arguments.add_argument(
        "--after-test", metavar="CMD", help="invoked after the test command"
    )

    tests = parser.add_argument_group(
        "TESTS",
        """\
        unittest\'s discover is used by default (if available)""",
    )
    tests.add_argument(
        "--test-nose",
        action="store_true",
        help="use nose module in --test step",
    )
    tests.add_argument(
        "--test-nose2",
        action="store_true",
        help="use nose2 module in --test step",
    )
    tests.add_argument(
        "--test-pytest",
        action="store_true",
        help="use pytest module in --test step",
    )
    tests.add_argument(
        "--test-tox",
        action="store_true",
        help="use tox in --test step",
    )
    tests.add_argument(
        "--test-stestr",
        action="store_true",
        help="use stestr in --test step",
    )
    tests.add_argument(
        "--test-unittest",
        action="store_true",
        help="use unittest in --test step",
    )
    tests.add_argument(
        "--test-custom",
        action="store_true",
        help="use custom command in --test step",
    )

    dirs = parser.add_argument_group("DIRECTORIES")
    dirs.add_argument(
        "-d",
        "--dir",
        action="store",
        metavar="DIR",
        help="source files directory - base for other relative dirs [default: CWD]",
    )
    dirs.add_argument(
        "--dest-dir",
        action="store",
        metavar="DIR",
        dest="destdir",
        help="destination directory [default: debian/tmp]",
    )
    dirs.add_argument(
        "--ext-dest-dir",
        action="store",
        metavar="DIR",
        dest="ext_destdir",
        help="destination directory for .so files",
    )
    dirs.add_argument(
        "--ext-pattern",
        action="store",
        metavar="PATTERN",
        help="regular expression for files that should be moved"
        " if --ext-dest-dir is set [default: .so files]",
    )
    dirs.add_argument(
        "--ext-sub-pattern",
        action="store",
        metavar="PATTERN",
        help="pattern to change --ext-pattern's filename or path",
    )
    dirs.add_argument(
        "--ext-sub-repl",
        action="store",
        metavar="PATTERN",
        help="replacement for match from --ext-sub-pattern," " empty string by default",
    )
    dirs.add_argument(
        "--install-dir",
        action="store",
        metavar="DIR",
        help="installation directory [default: .../dist-packages]",
    )
    dirs.add_argument(
        "--name",
        action="store",
        help="use this name to guess destination directories",
    )

    limit = parser.add_argument_group("LIMITATIONS")
    limit.add_argument(
        "-s",
        "--system",
        help="select a build system [default: auto-detection]",
    )
    limit.add_argument(
        "-p",
        "--pyver",
        action=AppendVersions,
        dest="versions",
        help="""build for Python VERSION.
                               This option can be used multiple times
                               [default: all supported Python 3.X versions]""",
    )
    limit.add_argument(
        "-i",
        "--interpreter",
        action="append",
        help="change interpreter [default: python{version}]",
    )
    limit.add_argument(
        "--disable", metavar="ITEMS", help="disable action, interpreter or version"
    )
    limit.add_argument(
        "-j",
        "--parallel",
        help="specify number of parallel jobs [default: 1]",
    )
    return parser



# Copyright © 2012-2020 Piotr Ożarowski <piotr@debian.org>
#           © 2020 Scott Kitterman <scott@kitterman.com>
#           © 2021 Stuart Prescott <stuart@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from contextlib import contextmanager
from pathlib import Path
from typing import override
from collections.abc import Iterator
import logging
import os.path as osp
import shutil
import sysconfig
import tomllib

try:
    from installer import install
    from installer.destinations import SchemeDictionaryDestination
    from installer.sources import WheelFile
    from installer.utils import parse_metadata_file
except ModuleNotFoundError:
    SchemeDictionaryDestination = WheelFile = install = parse_metadata_file = None  # type: ignore

from dhpython.build.base import Base, shell_command, create_setuptools_parallel_cfg
from dhpython.build.types import Context, Args
from dhpython.tools import dpkg_architecture

log = logging.getLogger('dhpython')


def _build_backend() -> str | None:
    """Retrieve the build-system from pyproject.toml, where possible"""
    try:
        with open('pyproject.toml', 'rb') as f:
            pyproject = tomllib.load(f)
        backend = pyproject.get('build-system', {}).get('build-backend')
        if backend is None:
            return None
        assert isinstance(backend, str)
        return backend
    except NameError:
        # No toml, no autdetection
        return None
    except FileNotFoundError:
        # Not a PEP517 package
        return None
    except tomllib.TOMLDecodeError as e:
        log.warning("pyproject.toml: TOMLDecodeError: %s", str(e))
        return None


class BuildSystem(Base):
    DESCRIPTION = 'Generic PEP517 build system'
    SUPPORTED_INTERPRETERS = {'python3', 'python{version}'}
    REQUIRED_FILES = ['pyproject.toml']
    OPTIONAL_FILES = {}
    CLEAN_FILES = Base.CLEAN_FILES | {'build'}

    @override
    def detect(self, context: Context) -> int:
        """Return certainty level that this plugin describes the right build
        system

        This method uses cls.{REQUIRED}_FILES (pyroject.toml) only; any
        other PEP517 compliant builder (such as the flit) builder should
        indicate higher specificity than this plugin.

        :return: 0 <= certainty <= 100
        :rtype: int
        """
        result = super().detect(context)
        # Temporarily reduce the threshold while we're in beta
        result -= 20

        if _build_backend():
            result += 10

        if result > 100:
            return 100
        return result

    @override
    def clean(self, context: Context, args: Args) -> None:
        super().clean(context, args)
        if osp.exists(args.interpreter.binary()):
            log.debug("removing '%s' (and everything under it)", args.build_dir)
            if osp.isdir(args.build_dir):
                shutil.rmtree(args.build_dir)

    @override
    def configure(self, context: Context, args: Args) -> None:
        if install is None:
            raise Exception("PEP517 plugin dependencies are not available. "
                            "Please Build-Depend on pybuild-plugin-pyproject.")
        # No separate configure step

    @override
    def build(self, context: Context, args: Args) -> None:
        self.build_wheel(context, args)
        self.unpack_wheel(context, args)

    def _backend_config_settings(self, args: Args) -> list[str]:
        backend = _build_backend()
        if backend == "mesonpy":
            arch_data = dpkg_architecture()
            return [
                # build_dir is where we unpack the wheel, so we need to build
                # elsewhere.
                f"build-dir={args.home_dir}/meson-build",
                "compile-args=--verbose",
                # From Debhelper's meson.pm
                "setup-args=--wrap-mode=nodownload",
                "setup-args=--prefix=/usr",
                "setup-args=--sysconfdir=/etc",
                "setup-args=--localstatedir=/var",
                "setup-args=--libdir=lib/" + arch_data["DEB_HOST_MULTIARCH"],
            ]
        return []

    @shell_command
    @create_setuptools_parallel_cfg(backend=_build_backend)
    @override
    def build_wheel(self, context: Context, args: Args) -> str:
        """ build a wheel using the PEP517 builder defined by upstream """
        log.info('Building wheel for %s with "build" module', args.interpreter)
        config_settings = self._backend_config_settings(args)
        context.ENV['FLIT_NO_NETWORK'] = '1'
        context.ENV['HOME'] = args.home_dir
        return ('{interpreter} -m build '
                '--skip-dependency-check --no-isolation --wheel '
                '--outdir ' + args.home_dir + ' ' +
                ' '.join(('--config-setting ' + setting)
                         for setting in config_settings) +
                ' {args}'
               )

    @contextmanager
    def opened_wheel(self, context: Context, args: Args) -> Iterator[WheelFile]:
        wheel_fn = self.built_wheel(context, args)
        if not wheel_fn:
            raise AssertionError("Unable to locate a built wheel")
        wheel = Path(wheel_fn)
        if wheel.name.startswith('UNKNOWN'):
            raise Exception(f'UNKNOWN wheel found: {wheel.name}. Does '
                            'pyproject.toml specify a build-backend?')
        with WheelFile.open(wheel) as source:
            yield source

    def unpack_wheel(self, context: Context, args: Args) -> None:
        """ unpack the wheel into pybuild's normal  """
        log.info('Unpacking wheel built for %s with "installer" module', args.interpreter)
        extras = {}
        for extra in ('scripts', 'data', 'include'):
            path = Path(args.home_dir) / extra
            if path.exists():
                log.warning('%s directory already exists, skipping unpack. '
                            'Is the Python package being built twice?',
                            extra.title())
                return
            extras[extra] = str(path)
        destination = SchemeDictionaryDestination(
            {
                'platlib': args.build_dir,
                'purelib': args.build_dir,
                'scripts': extras["scripts"],
                'data': extras["data"],
                'headers': extras["include"],
            },
            interpreter=args.interpreter.binary_dv,
            script_kind='posix',
        )

        with self.opened_wheel(context, args) as source:
            install(
                source=source,
                destination=destination,
                additional_metadata={},
            )

    @override
    def install(self, context: Context, args: Args) -> None:
        log.info('Copying package built for %s to destdir', args.interpreter)
        try:
            paths = sysconfig.get_paths(scheme='deb_system')
        except KeyError:
            # Debian hasn't patched sysconfig schemes until 3.10
            # TODO: Introduce a version check once sysconfig is patched.
            paths = sysconfig.get_paths(scheme='posix_prefix')

        # start by copying the data, scripts, and headers
        for extra in ('data', 'scripts', 'include'):
            src_dir = Path(args.home_dir) / extra
            if not src_dir.exists():
                continue
            if extra == 'include':
                with self.opened_wheel(context, args) as source:
                    metadata = parse_metadata_file(
                        source.read_dist_info("METADATA")
                    )
                    extra_path = osp.join(
                        args.interpreter.include_dir, metadata["Name"]
                    )
            else:
                extra_path = paths[extra]
            target_dir = args.destdir + extra_path
            log.debug('Copying %s directory contents from %s -> %s',
                      extra, src_dir, target_dir)
            shutil.copytree(
                src_dir,
                target_dir,
                dirs_exist_ok=True,
            )

        # then copy the modules
        module_dir = args.build_dir
        target_dir = args.destdir + args.install_dir
        log.debug('Copying module contents from %s -> %s',
                  module_dir, target_dir)
        shutil.copytree(
            module_dir,
            target_dir,
            dirs_exist_ok=True,
        )

    @shell_command
    @override
    def test(self, context: Context, args: Args) -> str:
        scripts = Path(args.home_dir) / 'scripts'
        if scripts.exists():
            context.ENV['PATH'] = f"{scripts}:{context.ENV['PATH']}"
        context.ENV['HOME'] = args.home_dir
        return super().test_cmd(context, args)



# Copyright © 2010-2020 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import email
import logging
import platform
import os
import re
import subprocess
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum, StrEnum, auto
from functools import cache, partial
from os.path import exists, isdir, join
from typing import Literal, NamedTuple, TypedDict, cast

if __name__ == "__main__":
    import sys

    sys.path.append(os.path.abspath(join(os.path.dirname(__file__), "..")))

from dhpython import (
    PKG_PREFIX_MAP,
    PUBLIC_DIR_RE,
    PYDIST_DIRS,
    PYDIST_OVERRIDES_FNAMES,
    PYDIST_DPKG_SEARCH_TPLS,
)
from dhpython.debhelper import BD
from dhpython.markers import ComplexEnvironmentMarker, parse_environment_marker
from dhpython.options import DHPythonOptions
from dhpython.version import get_requested_versions, Version

log = logging.getLogger("dhpython")

PYDIST_RE = re.compile(
    r"""
    (?P<name>[A-Za-z][A-Za-z0-9_.-]*)            # Python distribution name
    \s*
    (?P<vrange>(?:-?\d\.\d+(?:-(?:\d\.\d+)?)?)?) # version range
    \s*
    (?P<dependency>(?:[a-z][^;]*)?)              # Debian dependency
    (?:  # optional upstream version -> Debian version translator
        ;\s*
        (?P<standard>PEP386|PEP440)?             # PEP-386 / PEP-440 mode
        \s*
        (?P<rules>(?:s|tr|y).*)?                 # translator rules
    )?
    """,
    re.VERBOSE,
)
REQUIRES_RE = re.compile(
    r"""
    (?P<name>[A-Za-z][A-Za-z0-9_.-]*)    # Python distribution name
    \s*
    (?P<enabled_extras>(?:\[[^\]]*\])?)  # ignored for now
    \s*
    \(?  # optional parenthesis
    (?:  # optional minimum/maximum version
        (?P<operator><=?|>=?|==|!=|~=)
        \s*
        (?P<version>(\w|[-.*])+)
        (?:  # optional interval minimum/maximum version
            \s*
            ,
            \s*
            (?P<operator2><=?|>=?|==|!=)
            \s*
            (?P<version2>(\w|[-.])+)
        )?
    )?
    \)?  # optional closing parenthesis
    \s*
    (?:;  # optional environment markers
        (?P<environment_marker>.+)
    )?
    """,
    re.VERBOSE,
)
EXTRA_RE = re.compile(
    r"""
    ;
    \s*
    extra
    \s*
    ==
    \s*
    (?P<quote>['"])
    (?P<section>[a-zA-Z0-9-_.]+)
    (?P=quote)
    """,
    re.VERBOSE,
)
REQ_SECTIONS_RE = re.compile(
    r"""
    ^
    \[
    (?P<section>[a-zA-Z0-9-_.]+)?
    \s*
    (?::
        (?P<environment_marker>.+)
    )?
    \]
    \s*
    $
    """,
    re.VERBOSE,
)
DEB_VERS_OPS = {
    "==": "=",
    "<": "<<",
    ">": ">>",
    "~=": ">=",
}
# Optimize away any dependencies on Python less than:
MIN_PY_VERSION = [3, 11]


class Standard(StrEnum):
    PEP386 = "PEP386"
    PEP440 = "PEP440"


class ModificationAction(Enum):
    SKIP = auto()
    KEEP = auto()
    APPEND = auto()
    PREPEND = auto()


class RequirementModification(NamedTuple):
    action: ModificationAction
    alternative: str | None = None


@dataclass
class PyDist:
    name: str
    versions: set[Version]
    dependency: str
    standard: Standard | None
    rules: list[str]


def validate(fpath: str) -> bool:
    """Check if pydist file looks good."""
    with open(fpath, encoding="utf-8") as fp:
        for line in fp:
            line = line.strip()
            if line.startswith("#") or not line:
                continue
            if not PYDIST_RE.match(line):
                log.error(
                    "invalid pydist data in file %s: %s", fpath.rsplit("/", 1)[-1], line
                )
                return False
    return True


@cache
def load(impl: str) -> dict[str, list[PyDist]]:
    """Load information about installed Python distributions.

    :param impl: interpreter implementation, f.e. cpython3
    :type impl: str
    """
    fname = PYDIST_OVERRIDES_FNAMES.get(impl)
    if fname and exists(fname):
        to_check = [fname]  # first one!
    else:
        to_check = []

    dname = PYDIST_DIRS.get(impl)
    if dname and isdir(dname):
        to_check.extend(join(dname, i) for i in os.listdir(dname))

    fbdir = os.environ.get("DH_PYTHON_DIST", "/usr/share/dh-python/dist/")
    fbname = join(fbdir, f"{impl}_fallback")
    if exists(fbname):  # fall back generated at dh-python build time
        to_check.append(fbname)  # last one!

    result: dict[str, list[PyDist]] = {}
    for fpath in to_check:
        assert fpath
        with open(fpath, encoding="utf-8") as fp:
            for line in fp:
                line = line.strip()
                if line.startswith("#") or not line:
                    continue
                if not (m := PYDIST_RE.search(line)):
                    raise Exception(f"invalid pydist line: {line} (in {fpath})")
                data = m.groupdict()
                dist = PyDist(
                    name=normalize_name(data["name"]),
                    versions=get_requested_versions(impl, data["vrange"]),
                    dependency=data["dependency"].strip(),
                    rules=data["rules"].split(";") if data["rules"] else [],
                    standard=cast(Standard | None, data["standard"]),
                )
                result.setdefault(dist.name, []).append(dist)
    return result


def merge_alternative_dependency(
    first: str,
    second: str,
) -> str:
    return f"{first} | {second}"


def guess_dependency(
    impl: str,
    req: str,
    version: Version | str | None = None,
    bdep: BD | None = None,
    accept_upstream_versions: bool = False,
) -> str | None:
    bdep = bdep or {}
    log.debug("trying to find dependency for %s (python=%s)", req, version)
    if isinstance(version, str):
        version = Version(version)

    # some upstreams have weird ideas for distribution name...
    if not (m := re.compile(r"([^!><=~ \(\)\[;]+)(.*)").match(req)):
        raise Exception(f"Unable to parse requirement: {req}")
    name, rest = m.groups()
    # TODO: check stdlib and dist-packaged for name.py and name.so files
    req = normalize_name(name) + rest

    data = load(impl)
    if not (m := REQUIRES_RE.match(req)):
        log.info(
            "please ask dh_python3 author to fix REQUIRES_RE "
            "or your upstream author to fix requires.txt"
        )
        raise Exception(f"requirement is not valid: {req}")
    req_d: dict[str, str] = m.groupdict()

    merge_marker: Callable[[str], str] = str
    if req_d["environment_marker"]:
        modification = check_environment_marker_restrictions(
            req, req_d["environment_marker"], impl
        )
        match modification.action:
            case ModificationAction.SKIP:
                return None
            case ModificationAction.KEEP:
                pass
            case ModificationAction.APPEND:
                assert modification.alternative
                merge_marker = partial(
                    merge_alternative_dependency, second=modification.alternative
                )
            case ModificationAction.PREPEND:
                assert modification.alternative
                merge_marker = partial(
                    merge_alternative_dependency, modification.alternative
                )

    name = req_d["name"]
    details = data.get(normalize_name(name))
    if details:
        log.debug("dependency: module %s is known to pydist: %r", name, details)
        for item in details:
            if version and version not in item.versions:
                # rule doesn't match version, try next one
                continue
            if not item.dependency:
                log.debug("dependency: requirement ignored")
                return None  # this requirement should be ignored
            if item.dependency.endswith(")"):
                # no need to translate versions if version is hardcoded in
                # Debian dependency
                log.debug("dependency: requirement already has hardcoded version")
                return merge_marker(item.dependency)
            if req_d["operator"] == "==" and req_d["version"].endswith("*"):
                # Translate "== 1.*" to "~= 1.0"
                req_d["operator"] = "~="
                req_d["version"] = req_d["version"].replace("*", "0")
                log.debug("dependency: translated wildcard version to semver limit")
            if (
                req_d["version"]
                and (item.standard or item.rules)
                and req_d["operator"] not in (None, "!=")
            ):
                o = _translate_op(req_d["operator"])
                v = _translate(req_d["version"], item.rules, item.standard)
                if req_d["operator"] == "==" and req_d["operator2"] is None:
                    # Loosen for Debian revisions
                    m = re.search(r"(.*)(\d+)(\D*)$", v)
                    if m:
                        max_v = (
                            m.group(1) + str((int(m.group(2))) + 1) + m.group(3) + "~"
                        )
                    else:
                        max_v = v + ".0~"
                    d = (
                        merge_marker(f"{item.dependency} (>= {v})")
                        + ", "
                        + merge_marker(f"{item.dependency} (<< {max_v})")
                    )
                else:
                    d = merge_marker(f"{item.dependency} ({o} {v})")
                if req_d["version2"] and req_d["operator2"] not in (None, "!="):
                    o2 = _translate_op(req_d["operator2"])
                    v2 = _translate(req_d["version2"], item.rules, item.standard)
                    d += ", " + merge_marker(f"{item.dependency} ({o2} {v2})")
                elif req_d["operator"] == "~=":
                    o2 = "<<"
                    v2 = _translate(
                        _max_compatible(req_d["version"]),
                        item.rules,
                        item.standard,
                    )
                    d += ", " + merge_marker(f"{item.dependency} ({o2} {v2})")
                log.debug("dependency: constructed version")
                return d
            elif (
                accept_upstream_versions
                and req_d["version"]
                and req_d["operator"] not in (None, "!=")
            ):
                o = _translate_op(req_d["operator"])
                d = merge_marker(f"{item.dependency} ({o} {req_d['version']})")
                if req_d["version2"] and req_d["operator2"] not in (None, "!="):
                    o2 = _translate_op(req_d["operator2"])
                    d += ", " + merge_marker(
                        f"{item.dependency} ({o2} {req_d['version2']})"
                    )
                elif req_d["operator"] == "~=":
                    o2 = "<<"
                    d += ", " + merge_marker(
                        f"{item.dependency} ({o2} {_max_compatible(req_d['version'])})"
                    )
                log.debug("dependency: constructed upstream version")
                return d
            else:
                if item.dependency in bdep:
                    if None in bdep[item.dependency] and bdep[item.dependency][None]:
                        log.debug("dependency: included in build-deps with limits ")
                        return merge_marker(
                            f"{item.dependency} ({bdep[item.dependency][None]})"
                        )
                    # if arch in bdep[item.dependency]:
                    # TODO: handle architecture specific dependencies from build depends
                    #       (current architecture is needed here)
                log.debug("dependency: included in build-deps")
                return merge_marker(item.dependency)

    # search for Egg metadata file or directory (using dpkg -S)
    dpkg_query_tpl, regex_filter = PYDIST_DPKG_SEARCH_TPLS[impl]
    dpkg_query = dpkg_query_tpl.format(ci_regexp(dist_info_normalize_name(name)))

    log.debug("invoking dpkg -S %s", dpkg_query)
    process = subprocess.run(
        ("/usr/bin/dpkg", "-S", dpkg_query),
        check=False,
        encoding="UTF-8",
        capture_output=True,
    )
    if process.returncode == 0:
        result = set()
        for line in process.stdout.split("\n"):
            if not line.strip():
                continue
            pkg, path = line.split(":", 1)
            if regex_filter and not re.search(regex_filter, path):
                continue
            result.add(pkg)
        if len(result) > 1:
            log.error("more than one package name found for %s dist", name)
        elif not result:
            log.debug("dpkg -S did not find package for %s", name)
        else:
            log.debug("dependency: found a result with dpkg -S")
            return merge_marker(result.pop())
    else:
        log.debug("dpkg -S did not find package for %s: %s", name, process.stderr)

    pname = sensible_pname(impl, normalize_name(name))
    log.info(
        "Cannot find package that provides %s. "
        "Please add package that provides it to Build-Depends or "
        'add "%s %s" line to %s or add proper '
        "dependency to Depends by hand and ignore this info.",
        name,
        normalize_name(name),
        pname,
        PYDIST_OVERRIDES_FNAMES[impl],
    )
    return None


def check_environment_marker_restrictions(
    req: str, marker_str: str, impl: str
) -> RequirementModification:
    """Should we keep or skip a dependency based on its environment markers."""
    if impl != "cpython3":
        log.info("Ignoring environment markers for non-Python 3.x: %s", req)
        return RequirementModification(action=ModificationAction.SKIP)

    try:
        marker, op, value = parse_environment_marker(marker_str)
    except ComplexEnvironmentMarker:
        log.info("Ignoring complex environment marker: %s", req)
        return RequirementModification(action=ModificationAction.SKIP)

    # TODO: Use dynamic values when building arch-dependent
    # binaries, otherwise static values
    # TODO: Hurd values?
    supported_values = {
        "implementation_name": ("cpython", "pypy"),
        "os_name": ("posix",),
        "platform_system": ("Linux",),
        "platform_machine": (platform.machine(),),
        "platform_python_implementation": ("CPython", "PyPy"),
        "sys_platform": ("linux",),
    }
    if marker in supported_values:
        sv = supported_values[marker]
        if op in ("==", "!="):
            if op == "==" and value not in sv:
                log.debug("Skipping requirement (%s != %s): %s", value, sv, req)
                return RequirementModification(action=ModificationAction.SKIP)
            if op == "!=" and len(supported_values[marker]) > 1:
                log.debug(
                    (
                        "Keeping requirement as it applies to one of several "
                        "supported environments: %s"
                    ),
                    req,
                )
                return RequirementModification(action=ModificationAction.KEEP)
        else:
            log.info(
                "Skipping requirement with unhandled environment marker "
                "comparison: %s",
                req,
            )
            return RequirementModification(action=ModificationAction.SKIP)

    elif marker in ("python_version", "python_full_version", "implementation_version"):
        # TODO: Replace with full PEP-440 parser
        env_ver = value
        split_ver = value.split(".")
        if marker == "python_version":
            version_parts = 2
        elif marker == "python_full_version":
            version_parts = 3
        else:
            version_parts = len(split_ver)

        if "*" in env_ver:
            if split_ver.index("*") != len(split_ver) - 1:
                log.info("Skipping requirement with intermediate wildcard: %s", req)
                return RequirementModification(action=ModificationAction.SKIP)
            split_ver.pop()
            env_ver = ".".join(split_ver)
            if op == "==":
                if marker == "python_full_version":
                    marker = "python_version"
                    version_parts = 2
            elif op == "!=":
                if marker == "python_full_version":
                    marker = "python_version"
                    version_parts = 2
                else:
                    log.info(
                        "Ignoring wildcard != requirement, not "
                        "representable in Debian: %s",
                        req,
                    )
                    return RequirementModification(action=ModificationAction.KEEP)
            else:
                log.info("Skipping requirement with %s on a wildcard: %s", op, req)
                return RequirementModification(action=ModificationAction.SKIP)

        int_ver = []
        for ver_part in split_ver:
            if ver_part.isdigit():
                int_ver.append(int(ver_part))
            else:
                env_ver = ".".join(str(x) for x in int_ver)
                log.info(
                    "Truncating unparseable version %s to %s in %s", value, env_ver, req
                )
                break

        if len(int_ver) < version_parts:
            int_ver.append(0)
            env_ver += ".0"
        int_next_ver = int_ver.copy()
        int_next_ver[version_parts - 1] += 1
        next_ver = ".".join(str(x) for x in int_next_ver)
        # int_prev_ver = int_ver.copy()
        # int_prev_ver[version_parts - 1] -= 1
        # prev_ver = '.'.join(str(x) for x in int_prev_ver)

        # We try to do something somewhat sensible with micro versions
        # even though we don't currently emit them in python3-supported-min/max

        if op == "<":
            if int_ver <= MIN_PY_VERSION:
                return RequirementModification(action=ModificationAction.SKIP)
            return RequirementModification(
                action=ModificationAction.PREPEND,
                alternative=f"python3-supported-min (>= {env_ver})",
            )
        elif op == "<=":
            return RequirementModification(
                action=ModificationAction.PREPEND,
                alternative=f"python3-supported-min (>> {env_ver})",
            )
        elif op == ">=":
            if int_ver < MIN_PY_VERSION:
                return RequirementModification(action=ModificationAction.KEEP)
            return RequirementModification(
                action=ModificationAction.APPEND,
                alternative=f"python3-supported-max (<< {env_ver})",
            )
        elif op == ">":
            if int_ver < MIN_PY_VERSION:
                return RequirementModification(action=ModificationAction.KEEP)
            return RequirementModification(
                action=ModificationAction.APPEND,
                alternative=f"python3-supported-max (<= {env_ver})",
            )
        elif op == "==":
            if marker == "python_version":
                return RequirementModification(
                    action=ModificationAction.APPEND,
                    alternative=(
                        f"python3-supported-max (<< {env_ver}) "
                        f"| python3-supported-min (>= {next_ver})"
                    ),
                )
            return RequirementModification(
                action=ModificationAction.APPEND,
                alternative=(
                    f"python3-supported-max (<< {env_ver}) "
                    f"| python3-supported-min (>> {env_ver})"
                ),
            )
        elif op == "===":
            # === is arbitrary equality (PEP 440)
            if marker == "python_version":
                return RequirementModification(
                    action=ModificationAction.APPEND,
                    alternative=(
                        f"python3-supported-max (<< {env_ver}) "
                        f"| python3-supported-min (>> {env_ver})"
                    ),
                )
            else:
                log.info(
                    "Skipping requirement with %s environment marker, cannot "
                    "model in Debian deps: %s",
                    op,
                    req,
                )
                return RequirementModification(action=ModificationAction.SKIP)
        elif op == "~=":  # Compatible equality (PEP 440)
            int_ceq_next_ver = int_ver[:2]
            int_ceq_next_ver[1] += 1
            ceq_next_ver = ".".join(str(x) for x in int_ceq_next_ver)
            return RequirementModification(
                action=ModificationAction.APPEND,
                alternative=(
                    f"python3-supported-max (<< {env_ver}) "
                    f"| python3-supported-min (>= {ceq_next_ver})"
                ),
            )
        elif op == "!=":
            log.info(
                "Ignoring != comparison in environment marker, cannot "
                "model in Debian deps: %s",
                req,
            )
            return RequirementModification(action=ModificationAction.KEEP)

    elif marker == "extra":
        # Handled in section logic of parse_requires_dist()
        return RequirementModification(action=ModificationAction.KEEP)
    else:
        log.info("Skipping requirement with unknown environment marker: %s", marker)
        return RequirementModification(action=ModificationAction.SKIP)
    return RequirementModification(action=ModificationAction.KEEP)


class NewDependencies(TypedDict):
    depends: list[str]
    recommends: list[str]
    suggests: list[str]


def parse_pydep(
    impl: str,
    fname: str,
    bdep: BD | None = None,
    *,
    options: DHPythonOptions | None = None,
    depends_sec: list[str] | None = None,
    recommends_sec: list[str] | None = None,
    suggests_sec: list[str] | None = None,
) -> NewDependencies:
    depends_sec = depends_sec or []
    recommends_sec = recommends_sec or []
    suggests_sec = suggests_sec or []

    public_dir = PUBLIC_DIR_RE[impl].match(fname)
    ver = None
    if public_dir and public_dir.groups() and len(public_dir.group(1)) != 1:
        ver = public_dir.group(1)

    guess_deps = partial(
        guess_dependency,
        impl=impl,
        version=ver,
        bdep=bdep,
        accept_upstream_versions=getattr(options, "accept_upstream_versions", False),
    )

    result = NewDependencies(depends=[], recommends=[], suggests=[])
    modified: bool = False
    section: str | None = None
    modification = RequirementModification(action=ModificationAction.KEEP)
    processed = []
    with open(fname, encoding="utf-8") as fp:
        for line in fp:
            line = line.strip()
            if not line or line.startswith("#"):
                processed.append(line)
                continue
            if line.startswith("["):
                m = REQ_SECTIONS_RE.match(line)
                if not m:
                    log.info("Skipping section %s, unable to parse header", line)
                    processed.append(line)
                    # something that won't map to a result_key
                    section = "[non-existent section]"
                    continue
                section = m.group("section")
                modification = RequirementModification(action=ModificationAction.KEEP)
                if m.group("environment_marker"):
                    modification = check_environment_marker_restrictions(
                        line, m.group("environment_marker"), impl
                    )
                processed.append(line)
                continue
            result_key: Literal["depends", "recommends", "suggests"]
            if section:
                if section in depends_sec:
                    result_key = "depends"
                elif section in recommends_sec:
                    result_key = "recommends"
                elif section in suggests_sec:
                    result_key = "suggests"
                else:
                    processed.append(line)
                    continue
            else:
                result_key = "depends"

            dependency = None
            if modification.action != ModificationAction.SKIP:
                dependency = guess_deps(req=line)
            if dependency and modification.action == ModificationAction.APPEND:
                assert modification.alternative
                dependency = ", ".join(
                    part.strip() + " | " + modification.alternative
                    for part in dependency.split(",")
                )
            if dependency and modification.action == ModificationAction.PREPEND:
                assert modification.alternative
                dependency = ", ".join(
                    modification.alternative + " | " + part.strip()
                    for part in dependency.split(",")
                )

            if dependency:
                result[result_key].append(dependency)
                modified = True
            else:
                processed.append(line)
    if modified and public_dir:
        with open(fname, "w", encoding="utf-8") as fp:
            fp.writelines(i + "\n" for i in processed)
    return result


def parse_requires_dist(
    impl: str,
    fname: str,
    bdep: BD | None = None,
    *,
    options: DHPythonOptions | None = None,
    depends_sec: list[str] | None = None,
    recommends_sec: list[str] | None = None,
    suggests_sec: list[str] | None = None,
) -> NewDependencies:
    """Extract dependencies from a dist-info/METADATA file"""
    depends_sec = depends_sec or []
    recommends_sec = recommends_sec or []
    suggests_sec = suggests_sec or []

    public_dir = PUBLIC_DIR_RE[impl].match(fname)
    ver = None
    if public_dir and public_dir.groups() and len(public_dir.group(1)) != 1:
        ver = public_dir.group(1)

    guess_deps = partial(
        guess_dependency,
        impl=impl,
        version=ver,
        bdep=bdep,
        accept_upstream_versions=getattr(options, "accept_upstream_versions", False),
    )
    result = NewDependencies(depends=[], recommends=[], suggests=[])
    section = None
    with open(fname, encoding="utf-8") as fp:
        metadata = email.message_from_string(fp.read())
    requires = metadata.get_all("Requires-Dist", [])
    result_key: Literal["depends", "recommends", "suggests"]
    for req in requires:
        m = EXTRA_RE.search(req)
        result_key = "depends"
        if m:
            section = m.group("section")
            if section:
                if section in depends_sec:
                    result_key = "depends"
                elif section in recommends_sec:
                    result_key = "recommends"
                elif section in suggests_sec:
                    result_key = "suggests"
                else:
                    continue
        dependency = guess_deps(req=req)
        if dependency:
            result[result_key].append(dependency)
    return result


# https://packaging.python.org/en/latest/specifications/simple-repository-api/#normalized-names
def normalize_name(name: str) -> str:
    """Normalize a distribution name."""
    return re.sub(r"[-_.]+", "-", name).lower()


def dist_info_normalize_name(name: str) -> str:
    """Normalize a name for use in a .dist-info path."""
    return re.sub(r"[-_.]+", "_", name).lower()


def sensible_pname(impl: str, dist_name: str) -> str:
    """Guess Debian package name from normalized distribution name."""
    dist_name = dist_name.removeprefix("python-")
    return f"{PKG_PREFIX_MAP[impl]}-{dist_name}"


def ci_regexp(name: str) -> str:
    """Return case insensitive dpkg -S regexp."""
    return "".join(f"[{i.upper()}{i}]" if i.isalpha() else i for i in name.lower())


PEP386_PRE_VER_RE = re.compile(r"[-.]?(alpha|beta|rc|dev|a|b|c)")
PEP440_PRE_VER_RE = re.compile(r"[-.]?(a|b|rc)")
GROUP_RE = re.compile(r"\$(\d+)")


def _pl2py(pattern: str) -> str:
    r"""Convert Perl RE patterns used in uscan to Python's

    >>> print(_pl2py('foo$3'))
    foo\g<3>
    """
    return GROUP_RE.sub(r"\\g<\1>", pattern)


def _max_compatible(version: str) -> str:
    """Return the maximum version compatible with `version` in PEP440 terms,
    used by ~= requires version specifiers.

    https://www.python.org/dev/peps/pep-0440/#compatible-release

    >>> _max_compatible('2.2')
    '3'
    >>> _max_compatible('1.4.5')
    '1.5'
    >>> _max_compatible('1.3.alpha4')
    '2'
    >>> _max_compatible('2.1.3.post5')
    '2.2'

    """
    v = Version(version)
    v.serial = None
    v.releaselevel = None
    if v.micro is not None:
        v.micro = None
        return str(v + 1)
    v.minor = None
    return str(v + 1)


def _translate(version: str, rules: list[str], standard: Standard | None) -> str:
    """Translate Python version into Debian one.

    >>> _translate('1.C2betac', ['s/c//gi'], None)
    '1.2beta'
    >>> _translate('5-fooa1.2beta3-fooD',
    ...     ['s/^/1:/', 's/-foo//g', 's:([A-Z]):+$1:'], 'PEP386')
    '1:5~a1.2~beta3+D'
    >>> _translate('x.y.x.z', ['tr/xy/ab/', 'y,z,Z,'], None)
    'a.b.a.Z'
    """
    for rule in rules:
        # uscan supports s, tr and y operations
        if rule.startswith(("tr", "y")):
            # Note: no support for escaped separator in the pattern
            pos = 1 if rule.startswith("y") else 2
            tmp = rule[pos + 1 :].split(rule[pos])
            version = version.translate(str.maketrans(tmp[0], tmp[1]))
        elif rule.startswith("s"):
            # uscan supports: g, u and x flags
            tmp = rule[2:].split(rule[1])
            pattern = re.compile(tmp[0])
            count = 1
            if tmp[2:]:
                flags = tmp[2]
                if "g" in flags:
                    count = 0
                if "i" in flags:
                    pattern = re.compile(tmp[0], re.I)
            version = pattern.sub(_pl2py(tmp[1]), version, count)
        else:
            log.warning("unknown rule ignored: %s", rule)
    if standard == Standard.PEP386:
        version = PEP386_PRE_VER_RE.sub(r"~\g<1>", version)
    elif standard == Standard.PEP440:
        version = PEP440_PRE_VER_RE.sub(r"~\g<1>", version)
    return version


def _translate_op(operator: str) -> str:
    """Translate Python version operator into Debian one.

    >>> _translate_op('==')
    '='
    >>> _translate_op('<')
    '<<'
    >>> _translate_op('<=')
    '<='
    """
    return DEB_VERS_OPS.get(operator, operator)


if __name__ == "__main__":
    impl = os.environ.get("IMPL", "cpython3")
    for i in sys.argv[1:]:
        if os.path.isfile(i):
            try:
                print(", ".join(parse_pydep(impl, i)["depends"]))
            except Exception as err:
                log.error("%s: cannot guess (%s)", i, err)
        else:
            try:
                print(guess_dependency(impl, i) or "")
            except Exception as err:
                log.error("%s: cannot guess (%s)", i, err)
